/*
 * eID Chrome OS project.
 *
 * Copyright 2017-2018 e-Contract.be BVBA. All rights reserved.
 * e-Contract.be BVBA proprietary/confidential. Use is subject to license terms.
 */

"use strict";

/**
 * Main constructor. Runs on Chroms OS, Linux, Mac OS X, and Android via Cordova and bleat.
 * 
 * @class
 * @constructor
 */
function BeIDACR3901U() {
    this.connectedServer = null;
    this.longResponseBytesLeft = 0;
    this.longResponse = null;
    this.sessionKey = null;
    this.apduCallback = null;
    this.authenticateCallback = null;
    this.cardPresent = false;
    this.ui = null;
    this.cardPresentCallback = null;
    this.powerOnCallback = null;
    this.powerOffCallback = null;
    this.progressStatus = false;
    this.rndB = null;
    this.errorCallback = null;
    // next is used by background.js
    this.vendorId = "072f";
    this.productId = "b000";
}

var ATR_PATTERN = [0x3b, 0x98, 0x00, 0x40, 0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0xad, 0x13, 0x10];

var ATR_MASK = [0xff, 0xff, 0x00, 0xff, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xf0];

/**
 * Checks whether the given ATR belongs to the BeID.
 * @param {Uint8Array} atr the ATR.
 * @returns {boolean}
 */
BeIDACR3901U.prototype.isBeID = function (atr) {
    if (atr.length !== ATR_PATTERN.length) {
        return false;
    }
    for (var idx = 0; idx < ATR_PATTERN.length; idx++) {
        if ((atr[idx] & ATR_MASK[idx]) !== ATR_PATTERN[idx]) {
            return false;
        }
    }
    return true;
};

/**
 * Starts the connection towards the ACR3901U bluetooth smart card reader.
 * @param {Object} ui the UI object.
 * @param {function():void} callback invoked when the eID card is present.
 * @param {function(Error):void} errorCallback invoked when an error occurred.
 */
BeIDACR3901U.prototype.start = function (ui, callback, errorCallback) {
    var beid = this;
    beid.ui = ui;
    beid.errorCallback = errorCallback;
    beid.getServer()
        .then(function () {
            beid.authenticateCardReader(function () {
                console.log("authenticated");
                console.log("card present: " + beid.cardPresent);
                if (!beid.cardPresent) {
                    beid.cardPresentCallback = callback;
                    beid.ui.setStatusText("Please insert your eID card...");
                } else {
                    beid.ui.setStatusText();
                    callback();
                }
            });
        })
        .catch(function (error) {
            console.error("error occurred: " + error);
            errorCallback(new Error(error));
        });
};

/**
 * Converts an Uint8Array to hexadecimal string.
 * @param {Uint8Array} a the Uint8Array to convert.
 * @returns {string} the hexadecimal value.
 */
BeIDACR3901U.prototype.Uint8ArrayToHex = function (a) {
    var ret = "";
    for (var idx = 0; idx < a.length; idx++) {
        ret += (a[idx] < 16 ? "0" : "") + a[idx].toString(16);
    }
    return ret;
};

BeIDACR3901U.prototype.handleResponseNotifications = function (event) {
    var beid = this;
    console.log("response notification event received");
    console.dir(event.target);
    var value = event.target.value.buffer;
    var responseBytes = new Uint8Array(value);
    console.log("response bytes: " + responseBytes.length);
    console.log("response: " + beid.Uint8ArrayToHex(responseBytes));

    beid.progressStatus = !beid.progressStatus;
    beid.ui.progress(beid.progressStatus);

    if (beid.longResponseBytesLeft) {
        console.log("long response bytes left: " + beid.longResponseBytesLeft);
        var currentLongResponse = beid.longResponse;
        beid.longResponse = new Uint8Array(beid.longResponse.byteLength + responseBytes.byteLength);
        beid.longResponse.set(currentLongResponse, 0);
        beid.longResponse.set(responseBytes, currentLongResponse.byteLength);
        beid.longResponseBytesLeft -= responseBytes.byteLength;
        if (beid.longResponseBytesLeft) {
            console.log("expecting more bytes: " + beid.longResponseBytesLeft);
            return;
        }
    } else {
        var size = responseBytes[1] + responseBytes[2] * 256;
        console.log("size: " + size);
        beid.longResponse = responseBytes;
        if (size > 20 - 3) {
            console.log("long response mode");
            if (beid.sessionKey !== null) {
                beid.longResponseBytesLeft = size - 20 + 3;
                console.log("expecting more bytes: " + beid.longResponseBytesLeft);
                return;
            }
            else {
                console.warn("Ignoring long reponse mode for unauthenticated reader");
            }
        }
    }

    var checksum = beid.calculateChecksum(beid.longResponse);
    if (checksum !== beid.longResponse[beid.longResponse.length - 1]) {
        console.error("checksum incorrect");
        beid.ui.setStatusText("checksum error", "ERROR");
        beid.close();
        beid.errorCallback(new Error("checksum error"));
        return;
    }

    if (beid.longResponse[0] === 0x20) {
        // RDR_to_SPH_AuthRsp1
        console.log("authentication response received");
        var key = [0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff];
        var encryptedRndA = beid.longResponse.subarray(3, 3 + 16);
        console.log("encrypted RND_A: " + beid.Uint8ArrayToHex(encryptedRndA));
        console.log("encrypted RND_A length: " + encryptedRndA.slice().length);
        var iv = [0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00];
        var decipher = forge.cipher.createDecipher("AES-CBC", key);
        decipher.mode.unpad = false;
        decipher.start({ iv: iv });
        decipher.update(new forge.util.ByteBuffer(new Uint8Array(encryptedRndA.slice().buffer)));
        var result = decipher.finish();
        console.log("decipher result: " + result);
        var rndAByteStringBuffer = decipher.output;
        console.log("RND_A: " + rndAByteStringBuffer.toHex());
        console.log("RND_A size: %s", rndAByteStringBuffer.length());
        var rndA = forge.util.binary.raw.decode(rndAByteStringBuffer.getBytes());
        beid.rndB = new Uint8Array(16);
        window.crypto.getRandomValues(beid.rndB);
        console.log("rndB: " + beid.Uint8ArrayToHex(beid.rndB));
        var rndC = new Uint8Array(32);
        rndC.set(rndA, 16);
        rndC.set(beid.rndB, 0);
        console.log("rndC: " + beid.Uint8ArrayToHex(rndC.slice()));

        beid.sessionKey = new Uint8Array(16);
        beid.sessionKey.set(rndA.subarray(0, 8), 0);
        beid.sessionKey.set(beid.rndB.subarray(0, 8), 8);
        console.log("session key: " + beid.Uint8ArrayToHex(beid.sessionKey));

        decipher = forge.cipher.createDecipher("AES-CBC", key);
        decipher.mode.unpad = false;
        decipher.start({ iv: iv });
        decipher.update(new forge.util.ByteBuffer(new Uint8Array(rndC.slice().buffer)));
        result = decipher.finish();
        console.log("decipher result: " + result);
        var encryptedRndCByteStringBuffer = decipher.output;
        //console.log("encrypted RND_C size: %s", encryptedRndCByteStringBuffer.length());
        var encryptedRndC = forge.util.binary.raw.decode(encryptedRndCByteStringBuffer.getBytes());
        console.log("encrypted rndC: " + beid.Uint8ArrayToHex(encryptedRndC));
        console.log("encrypted RND_C size: " + encryptedRndC.byteLength);

        //SPH_to_RDR_AuthRsp
        var authResp = new Uint8Array(36);
        authResp[0] = 0x71;
        authResp[1] = 0x21;
        authResp[2] = 0x00;
        authResp.set(encryptedRndC.slice(), 3);
        beid.setChecksum(authResp);
        console.log("auth resp: " + beid.Uint8ArrayToHex(authResp));
        console.log("auth resp size: " + authResp.byteLength);

        beid.writeData(authResp);
    } else if (beid.longResponse[0] === 0x21) {
        // RDR_to_SPH_AuthRsp2
        var size = beid.longResponse[1] + beid.longResponse[2] * 256;
        var encryptedRndB2 = beid.longResponse.subarray(3, 3 + size - 1);
        var key = [0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff];
        var iv = [0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00];
        var decipher = forge.cipher.createDecipher("AES-CBC", key);
        decipher.mode.unpad = false;
        decipher.start({ iv: iv });
        decipher.update(new forge.util.ByteBuffer(new Uint8Array(encryptedRndB2.slice().buffer)));
        var result = decipher.finish();
        console.log("decipher result: " + result);
        var rndB2 = forge.util.binary.raw.decode(decipher.output.getBytes());
        console.log("RND_B 2: " + beid.Uint8ArrayToHex(rndB2));
        if (beid.rndB.toString() === rndB2.toString()) {
            console.log("successful authentication performed");
            // we cannot depend on cardStatusNotification here to already have arrived, so we check for card presence manually too.
            //beid.authenticateCallback();
            var getCardPresenceReq = new Uint8Array([0x65, 0x01, 0x00, 0x64]);
            beid.dataRequest(getCardPresenceReq);
        } else {
            console.error("RND_B mismatch");
            beid.ui.setStatusText("RND_B mismatch", "ERROR");
            beid.close();
            beid.errorCallback(new Error("RND_B mismatch"));
        }
    } else if (beid.longResponse[0] === 0x22) {
        // RDR_to_SPH_DataRsp
        console.log("data response received");
        if (beid.sessionKey === null) {
            console.warn("Discarding data response before authentication success");
            return;
        }
        console.log("response: " + beid.Uint8ArrayToHex(beid.longResponse));
        var size = beid.longResponse[1] + beid.longResponse[2] * 256;
        console.log("encrypted data size: " + (size - 1));
        var encryptedData = beid.longResponse.subarray(3, 3 + size - 1);
        var key = [0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff];
        var iv = [0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00];
        var decipher = forge.cipher.createDecipher("AES-CBC", Array.from(beid.sessionKey));
        decipher.mode.unpad = false;
        decipher.start({ iv: iv });
        decipher.update(new forge.util.ByteBuffer(new Uint8Array(encryptedData.slice().buffer)));
        var result = decipher.finish();
        console.log("decipher result: " + result);
        var decryptedData = forge.util.binary.raw.decode(decipher.output.getBytes());
        console.log("decrypted data: " + beid.Uint8ArrayToHex(decryptedData));
        if (decryptedData[0] === 0x12) {
            // Card Power On response
            var atrSize = decryptedData[1] + decryptedData[2] * 256 - 1;
            console.log("ATR size: " + atrSize);
            var atr = decryptedData.subarray(3, 3 + atrSize);
            console.log("ATR: " + beid.Uint8ArrayToHex(atr));
            if (beid.isBeID(atr)) {
                beid.powerOnCallback(atr);
            } else {
                console.error("Not an eID card.");
                beid.ui.setStatusText("Not an eID card.", "ERROR");
                beid.ui.reset();
                beid.close();
                beid.errorCallback(new Error("Not an eID card."));
            }
        } else if (decryptedData[0] === 0x13) {
            console.log("card power off done.");
            beid.powerOffCallback();
        } else if (decryptedData[0] === 0x11) {
            console.log("APDU response received");
            var apduSize = decryptedData[1] + decryptedData[2] * 256 - 1;
            var apduResponse = decryptedData.subarray(3, 3 + apduSize);
            console.log("APDU response: " + beid.Uint8ArrayToHex(apduResponse));
            beid.apduCallback(apduResponse);
        } else if (decryptedData[0] === 0x91 || decryptedData[0] === 0x92) {
            var errorCode = decryptedData[3];
            var errorMessage = "unknown error";
            if (errorCode === 0x05) {
                errorMessage = "card operation error";
            }
            beid.ui.setStatusText(errorMessage, "ERROR");
            // close here, cannot continue anyway
            beid.close();
            beid.errorCallback(new Error(errorMessage));
        } else if (decryptedData[0] === 0x14) {
            // Get Card Presence response
            var status = decryptedData[3];
            switch (status) {
                case 0x00:
                case 0x01:
                    beid.cardPresent = false;
                    break;
                case 0x02:
                case 0x03:
                    beid.cardPresent = true;
                    break;
            }
            // now that we have an updated card presence status, we can continue
            beid.authenticateCallback();
        } else {
            console.error("unknown data response: " + beid.Uint8ArrayToHex(decryptedData));
        }
    } else {
        console.error("unknown response: " + beid.Uint8ArrayToHex(longResponse));
    }
};

BeIDACR3901U.prototype.registerResponseNotificationHandler = function (server) {
    var beid = this;
    console.log("register response notification handler");
    var promise = new Promise(function (resolve) {
        server.getPrimaryService("ae448001-6e87-df33-adb8-51dfd29be725")
            .then(function (service) {
                return service.getCharacteristic("ae448002-6e87-df33-adb8-51dfd29be725");
            })
            .then(function (characteristic) {
                // we keep track of the registered instance for unregistration later on
                beid.handleResponseNotificationsInstance = function (event) {
                    beid.handleResponseNotifications(event);
                };
                characteristic.addEventListener('characteristicvaluechanged',
                    beid.handleResponseNotificationsInstance);
                characteristic.startNotifications().then(function () {
                    console.log('notification handling registered');
                    resolve(server);
                });
            });
    });
    return promise;
};

BeIDACR3901U.prototype.handleBatteryLevelNotifications = function (event) {
    var beid = this;
    console.log("battery level notification event received");
    var value = event.target.value;
    console.dir(value);
    var level = value.getUint8();
    console.log("level: " + level);
    switch (level) {
        case 0xfe:
            console.log("sufficient battery");
            beid.displayBatteryStatus(100);
            break;
        case 0x00:
            console.log("no battery");
            beid.displayBatteryStatus(0);
            break;
        case 0xff:
            console.log("USB modus");
            beid.displayBatteryStatus(0);
            break;
        default:
            console.log("low battery");
            beid.displayBatteryStatus(25);
            break;
    }
};

BeIDACR3901U.prototype.displayBatteryStatus = function (level) {
    var beid = this;
    if (typeof beid.ui.batteryStatus === "function") {
        beid.ui.batteryStatus(level);
    }
};

BeIDACR3901U.prototype.registerBatteryLevelNotificationHandler = function (server) {
    console.log("register battery level notification handler");
    var beid = this;
    var promise = new Promise(function (resolve) {
        server.getPrimaryService("ae44180f-6e87-df33-adb8-51dfd29be725")
            .then(function (service) {
                return service.getCharacteristic("ae442a19-6e87-df33-adb8-51dfd29be725");
            })
            .then(function (characteristic) {
                // we keep track of the registered instance for unregistration later on
                beid.handleBatteryLevelNotificationsInstance = function (event) {
                    beid.handleBatteryLevelNotifications(event);
                };
                characteristic.addEventListener('characteristicvaluechanged',
                    beid.handleBatteryLevelNotificationsInstance);
                characteristic.startNotifications().then(function () {
                    console.log('notification handling registered');
                    resolve(server);
                });
            });
    });
    return promise;
};

BeIDACR3901U.prototype.handleCardStatusNotifications = function (event) {
    var beid = this;
    console.log("card status notification event received");
    var value = event.target.value;
    var status = value.getUint8(1);
    switch (status) {
        case 0x02:
            console.log("no card present");
            beid.cardPresent = false;
            break;
        case 0x03:
            console.log("card present");
            beid.cardPresent = true;
            if (beid.ui !== null) {
                beid.ui.setStatusText();
            }
            if (beid.cardPresentCallback !== null) {
                var callback = beid.cardPresentCallback;
                beid.cardPresentCallback = null;
                callback();
            }
            break;
        case 0x04:
            console.log("reader has entered sleep mode");
            break;
    }
};

BeIDACR3901U.prototype.registerCardStatusNotificationHandler = function (server) {
    var beid = this;
    console.log("register card status notification handler");
    var promise = new Promise(function (resolve) {
        server.getPrimaryService("ae448001-6e87-df33-adb8-51dfd29be725")
            .then(function (service) {
                return service.getCharacteristic("ae448004-6e87-df33-adb8-51dfd29be725");
            })
            .then(function (characteristic) {
                // we keep track of the registered instance for unregistration later on
                beid.handleCardStatusNotificationsInstance = function (event) {
                    beid.handleCardStatusNotifications(event);
                };
                characteristic.addEventListener('characteristicvaluechanged', beid.handleCardStatusNotificationsInstance);
                characteristic.startNotifications().then(function () {
                    console.log('notification handling registered');
                    resolve(server);
                });
            });
    });
    return promise;
};

BeIDACR3901U.prototype.verify = function (applicationUrl, verifyCallback) {
    var beid = this;
    beid.ui.getPIN(applicationUrl, function (pin) {
        var verifyData = [0x20 | pin.length,
            0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF
        ];
        for (var idx = 0; idx < pin.length; idx += 2) {
            var digit1 = pin[idx];
            var digit2;
            if (idx + 1 < pin.length) {
                digit2 = pin[idx + 1];
            } else {
                digit2 = '0' + 0xf;
            }
            var value = ((digit1 - '0' << 4) + (digit2 - '0'));
            verifyData[idx / 2 + 1] = value;
        }
        // MVP:VERIFY
        var verifyApdu = [0x00, // cla
            0x20, // ins
            0x00, // p1
            0x01, // p2
            verifyData.length
        ].concat(verifyData);
        console.log("VERIFY APDU (direct)");
        beid.sendAPDU(verifyApdu, function (verifyApduResponse) {
            var verifySW = verifyApduResponse.subarray(verifyApduResponse.length - 2);
            if (verifySW[0] === 0x90 && verifySW[1] === 0x00) {
                beid.ui.setRetriesLeft();
                verifyCallback();
            } else if (verifySW[0] === 0x63) {
                var retriesLeft = verifySW[1] & 0x0f;
                beid.ui.setRetriesLeft(retriesLeft);
                beid.verify(applicationUrl, verifyCallback);
            } else if (verifySW[0] === 0x69 && verifySW[1] === 0x83) {
                beid.ui.setStatusText("CARD_BLOCKED", "WARNING");
                beid.close(beid.errorCallback, new Error("card blocked"));
            } else {
                console.error("Error SW %s", beid.Uint8ArrayToHex(verifySW));
                beid.ui.setStatusText("ERROR", "ERROR");
                beid.close(beid.errorCallback, new Error("error"));
            }
        });
    });
};

BeIDACR3901U.prototype.getResponse = function (dataSize, callback) {
    var beid = this;
    var getResponseApdu = [
        0x00, // CLA
        0xC0, // INS
        0x00, // P1
        0x00, // P2
        dataSize // Le
    ];
    console.log("GET RESPONSE");
    beid.sendAPDU(getResponseApdu, function (getResponseApduResponse) {
        var sw = getResponseApduResponse.subarray(getResponseApduResponse.length - 2);
        if (sw[0] !== 0x90 || sw[1] !== 0x00) {
            console.error("status word failure: %s", beid.Uint8ArrayToHex(sw));
            beid.ui.setStatusText("ERROR", "ERROR");
            beid.close();
            beid.errorCallback(new Error("status word failed"));
            return;
        }
        var data = getResponseApduResponse.subarray(0, getResponseApduResponse.length - 2);
        callback(data);
    });
};


BeIDACR3901U.prototype._cds = function (cdsApdu, applicationUrl, signatureCallback) {
    var beid = this;
    console.log("PSO:CDS APDU");
    beid.sendAPDU(cdsApdu, function (psoCdsResponse) {
        var psoCdsSW = psoCdsResponse.subarray(psoCdsResponse.length - 2);

        if (psoCdsSW[0] === 0x69 && psoCdsSW[1] === 0x82) {
            console.log("PIN required");
            beid.verify(applicationUrl, function () {
                beid._cds(cdsApdu, applicationUrl, signatureCallback);
            });
        } else if (psoCdsSW[0] === 0x61) {
            // GET RESPONSE required
            var signatureSize = psoCdsSW[1];
            beid.getResponse(signatureSize, function (data) {
                beid.ui.setStatusText("DONE", "SUCCESS");
                signatureCallback(data);
            });
        } else if (psoCdsSW[0] === 0x90 && psoCdsSW[1] === 0x00) {
            // some card readers can directly return the signature
            beid.ui.setStatusText("DONE", "SUCCESS");
            signatureCallback(psoCdsResponse.subarray(0, psoCdsResponse.length - 2));
        } else {
            console.error("Error SW %s", beid.Uint8ArrayToHex(psoCdsSW));
            beid.ui.setStatusText("ERROR", "ERROR");
            beid.close();
            beid.errorCallback(new Error("error"));
        }
    });
};

BeIDACR3901U.prototype.authenticate = function (challenge, applicationUrl, authenticationCallback) {
    var beid = this;
    // MSE:SET APDU
    console.log("MSE:SET APDU");
    beid.ui.setStatusText("AUTHENTICATING", "INFO");
    beid.sendAPDU([0x00, // cla
        0x22, // ins: Manage Security Environment
        0x41, // p1: Set the signature mode
        0xB6, // p2: value of Digital Signature Template in data field
        0x05, // Lc
        0x04, // length of data structure
        0x80, // algo ref
        0x02, // pkcs#1 sha1
        0x84, // tag for private key ref
        0x82 // auth key id
    ], function (mseSetResponse) {
        var mseSetSW = mseSetResponse.subarray(mseSetResponse.length - 2);
        if (mseSetSW[0] !== 0x90 || mseSetSW[1] !== 0x00) {
            console.error("MSE:SET status word: " + beid.Uint8ArrayToHex(mseSetSW));
            beid.ui.setStatusText("MSE:SET status word error", "ERROR");
            beid.close();
            beid.errorCallback(new Error("MSE:SET status word error"));
            return;
        }

        var nonce = new Uint8Array(20);
        window.crypto.getRandomValues(nonce);

        var encodedApplicationUrl = new Uint8Array(applicationUrl.split("").map(function (c) {
            return c.charCodeAt(0);
        }));
        var tbs = new Uint8Array(20 + 20 + encodedApplicationUrl.length);
        tbs.set(nonce, 0);
        tbs.set(challenge, 20);
        tbs.set(encodedApplicationUrl, 20 + 20);

        window.crypto.subtle.digest({
            name: "SHA-1"
        }, tbs).then(function (hash) {
            var hashView = new DataView(hash);
            console.log("hash length: " + hashView.byteLength);
            // PSO:CDS APDU
            var cdsApdu = [0x00, // cla
                0x2A, // ins: Perform Security Operation
                0x9E, // p1: PSO:CDS Compute Digital Signature
                0x9A, // p2
                hashView.byteLength // Lc
            ];
            for (var idx = 0; idx < hashView.byteLength; idx++) {
                cdsApdu.push(hashView.getUint8(idx));
            }
            cdsApdu.push(0x00); // Le
            beid._cds(cdsApdu, applicationUrl, function (signature) {
                console.log("signature size: " + (signature.length * 8) + " bit");
                authenticationCallback(nonce, signature);
            });
        });
    });
};

BeIDACR3901U.prototype.createDigestInfo = function (digestValue, digestAlgo) {
    var digestInfo;
    switch (digestAlgo) {
        case "SHA1":
        case "SHA-1":
            digestInfo = [0x30, 0x21, 0x30, 0x09, 0x06, 0x05, 0x2b, 0x0e, 0x03,
                0x02, 0x1a, 0x05, 0x00, 0x04, 0x14];
            break;
        case "SHA-224":
            digestInfo = [0x30, 0x2d, 0x30, 0x0d, 0x06, 0x09, 0x60, 0x86,
                0x48, 0x01, 0x65, 0x03, 0x04, 0x02, 0x04, 0x05, 0x00, 0x04, 0x1c];
            break;
        case "SHA256":
        case "SHA-256":
            digestInfo = [0x30, 0x31, 0x30, 0x0d, 0x06, 0x09, 0x60, 0x86,
                0x48, 0x01, 0x65, 0x03, 0x04, 0x02, 0x01, 0x05, 0x00, 0x04, 0x20];
            break;
        case "SHA384":
        case "SHA-384":
            digestInfo = [0x30, 0x41, 0x30, 0x0d, 0x06, 0x09, 0x60, 0x86,
                0x48, 0x01, 0x65, 0x03, 0x04, 0x02, 0x02, 0x05, 0x00, 0x04, 0x30];
            break;
        case "SHA512":
        case "SHA-512":
            digestInfo = [0x30, 0x51, 0x30, 0x0d, 0x06, 0x09, 0x60, 0x86,
                0x48, 0x01, 0x65, 0x03, 0x04, 0x02, 0x03, 0x05, 0x00, 0x04, 0x40];
            break;
        case "RIPEMD-128":
            digestInfo = [0x30, 0x1d, 0x30, 0x09, 0x06, 0x05, 0x2b, 0x24, 0x03,
                0x02, 0x02, 0x05, 0x00, 0x04, 0x10];
            break;
        case "RIPEMD-160":
            digestInfo = [0x30, 0x21, 0x30, 0x09, 0x06, 0x05, 0x2b, 0x24, 0x03,
                0x02, 0x01, 0x05, 0x00, 0x04, 0x14];
            break;
        case "RIPEMD-256":
            digestInfo = [0x30, 0x2d, 0x30, 0x09, 0x06, 0x05, 0x2b, 0x24, 0x03,
                0x02, 0x03, 0x05, 0x00, 0x04, 0x20];
            break;
        case "MD5_SHA1":
            digestInfo = [];
            break;
        default:
            console.log("unknown digest algo: " + digestAlgo);
    }
    for (var idx = 0; idx < digestValue.length; idx++) {
        digestInfo.push(digestValue[idx]);
    }
    return digestInfo;
};

BeIDACR3901U.prototype.sign = function (digestValue, digestAlgo, applicationUrl, signCallback) {
    var beid = this;
    beid.ui.setStatusText("SIGNING", "INFO");
    // MSE:SET APDU
    beid.sendAPDU([0x00, // cla
        0x22, // ins: Manage Security Environment
        0x41, // p1: Set the signature mode
        0xB6, // p2: value of Digital Signature Template in data field
        0x05, // Lc
        0x04, // length of data structure
        0x80, // algo ref
        0x01, // pkcs#1
        0x84, // tag for private key ref
        0x83 // non-repudiation key id
    ], function (mseSetResponse) {
        var mseSetSW = mseSetResponse.subarray(mseSetResponse.length - 2);
        if (mseSetSW[0] !== 0x90 || mseSetSW[1] !== 0x00) {
            console.error("MSE:SET status word: " + beid.Uint8ArrayToHex(mseSetSW));
            beid.ui.setStatusText("ERROR", "ERROR");
            beid.close();
            beid.errorCallback(new Error("status word error"));
            return;
        }
        console.log("digest value: " + digestValue);
        var digestInfo = beid.createDigestInfo(digestValue, digestAlgo);
        var cdsApdu = [0x00, // cla
            0x2A, // ins: Perform Security Operation
            0x9E, // p1: PSO:CDS Compute Digital Signature
            0x9A, // p2
            digestInfo.length // Lc
        ].concat(digestInfo);
        cdsApdu.push(0x00); // Le
        // for non-repudiation, we know we need a PIN verify anyway.
        beid.verify(applicationUrl, function () {
            beid._cds(cdsApdu, applicationUrl, function (signature) {
                console.log("signature size: " + (signature.length * 8) + " bit");
                signCallback(signature);
            });
        });
    });
};

/**
 * Authenticate the card reader. We cannot use a Promise here as we need to resolve outside the scoope of this function.
 */
BeIDACR3901U.prototype.authenticateCardReader = function (callback) {
    console.log("authenticate...");
    var beid = this;
    if (beid.sessionKey) {
        console.log("already authenticated");
        callback();
        return;
    }
    beid.ui.setStatusText("Authenticating card reader...", "INFO");
    beid.authenticateCallback = callback;
    beid.connectedServer.getPrimaryService("ae448001-6e87-df33-adb8-51dfd29be725")
        .then(function (service) {
            return service.getCharacteristic("ae448003-6e87-df33-adb8-51dfd29be725");
        })
        .then(function (characteristic) {
            // SPH_to_RDR_ReqAuth
            var authReq = new Uint8Array([0x70, 0x01, 0x00, 0x71]);
            console.log("request size: " + authReq.slice().buffer.byteLength);
            characteristic.writeValue(authReq.slice().buffer)
                .then(function () {
                    console.log("authentication ReqAuth writting...")
                })
                .catch(function (error) {
                    console.log("error writing: " + error);
                });
        });
};

BeIDACR3901U.prototype.getSessionKey = function () {
    var beid = this;
    var promise = new Promise(function (resolve) {
        if (beid.sessionKey) {
            console.log("using cached session key");
            return resolve(beid.sessionKey);
        }
        beid.getServer().then(function (server) {
            console.log("getServer promise resolved");
            return resolve(beid.sessionKey);
        });
    });
    return promise;
};

BeIDACR3901U.prototype.getServer = function () {
    var beid = this;
    var promise = new Promise(function (resolve, reject) {
        if (beid.connectedServer) {
            console.log("using cached connected server");
            return resolve(beid.connectedServer);
        }
        var bluetoothAPI;
        if (typeof bleat !== "undefined") {
            console.log("using bleat");
            bluetoothAPI = bleat;
        } else {
            console.log("using navigator.bluetooth");
            bluetoothAPI = navigator.bluetooth;
        }
        bluetoothAPI.requestDevice({
            "filters": [
                { "namePrefix": "ACR3901U" }
            ],
            "optionalServices": [
                "00001801-0000-1000-8000-00805f9b34fb",
                "ae448001-6e87-df33-adb8-51dfd29be725",
                "ae44180f-6e87-df33-adb8-51dfd29be725",
                "ae44180a-6e87-df33-adb8-51dfd29be725"
            ]
        })
            .then(function (device) {
                console.log("device: " + device.name);
                console.log("connected: " + device.gatt.connected);
                // we keep track of the registered instance for unregistration later on
                beid.serverDisconnectedHandlerInstance = function (event) {
                    beid.serverDisconnectedHandler(event);
                };
                device.addEventListener('gattserverdisconnected', beid.serverDisconnectedHandlerInstance);
                return device.gatt.connect();
            })
            .then(function (server) {
                return beid.registerCardStatusNotificationHandler(server);
            })
            .then(function (server) {
                return beid.registerBatteryLevelNotificationHandler(server);
            })
            .then(function (server) {
                return beid.registerResponseNotificationHandler(server);
            })
            .then(function (server) {
                beid.connectedServer = server;
                beid.authenticateCardReader(function () {
                    return resolve(beid.connectedServer);
                });
            })
            .catch(function (error) {
                console.error("error requesting device: " + error);
                reject("error requesting device: " + error);
            });
    });
    return promise;
};

BeIDACR3901U.prototype.serverDisconnectedHandler = function (event) {
    var beid = this;
    var device = event.target;
    console.log('Device ' + device.name + ' is disconnected.');
    if (beid.connectedServer === null) {
        // we're in the close() phase here
        return;
    }
    beid.connectedServer = null;
    beid.sessionKey = null;
    beid.ui.setStatusText("Card reader disconnected.", "ERROR");
    beid.errorCallback(new Error("disconnected"));
    beid.close();
    beid.ui.reset();
};

BeIDACR3901U.prototype.padToBlocksize = function (data) {
    console.log("data size: " + data.byteLength);
    var paddedSize = (data.byteLength + 15) / 16 << 4;
    console.log("padded size: " + paddedSize);
    var paddedArray = new Uint8Array(paddedSize);
    paddedArray.set(data);
    for (var idx = data.byteLength; idx < paddedArray.byteLength; idx++) {
        paddedArray[idx] = 0xff;
    }
    return paddedArray;
};

BeIDACR3901U.prototype.cardPowerOn = function (callback) {
    var beid = this;
    var cardPowerOn = new Uint8Array([0x62, 0x01, 0x00, 0x63]);
    beid.powerOnCallback = callback;
    beid.dataRequest(cardPowerOn);
};

BeIDACR3901U.prototype.cardPowerOff = function (callback) {
    var beid = this;
    var cardPowerOff = new Uint8Array([0x63, 0x01, 0x00, 0x62]);
    beid.powerOffCallback = callback;
    beid.dataRequest(cardPowerOff);
};

BeIDACR3901U.prototype.powerOffAndClose = function (callback) {
    var beid = this;
    beid.cardPowerOff(function () {
        beid.close(callback);
    });
};

BeIDACR3901U.prototype.unregisterCardStatusNotificationHandler = function () {
    var beid = this;
    var promise = new Promise(function (resolve) {
        if (beid.connectedServer === null) {
            resolve();
            return;
        }
        beid.connectedServer.getPrimaryService("ae448001-6e87-df33-adb8-51dfd29be725")
            .then(function (service) {
                return service.getCharacteristic("ae448004-6e87-df33-adb8-51dfd29be725");
            })
            .then(function (characteristic) {
                characteristic.removeEventListener('characteristicvaluechanged', beid.handleCardStatusNotificationsInstance);
                beid.handleCardStatusNotificationsInstance = null;
                characteristic.stopNotifications().then(function () {
                    resolve();
                });
            });
    });
    return promise;
};

BeIDACR3901U.prototype.unregisterBatteryLevelNotificationHandler = function () {
    var beid = this;
    var promise = new Promise(function (resolve) {
        if (beid.connectedServer === null) {
            resolve();
            return;
        }
        beid.connectedServer.getPrimaryService("ae44180f-6e87-df33-adb8-51dfd29be725")
            .then(function (service) {
                return service.getCharacteristic("ae442a19-6e87-df33-adb8-51dfd29be725");
            })
            .then(function (characteristic) {
                characteristic.removeEventListener('characteristicvaluechanged', beid.handleBatteryLevelNotificationsInstance);
                beid.handleBatteryLevelNotificationsInstance = null;
                characteristic.stopNotifications().then(function () {
                    resolve();
                });
            });
    });
    return promise;
};

BeIDACR3901U.prototype.unregisterResponseNotificationHandler = function () {
    var beid = this;
    var promise = new Promise(function (resolve) {
        if (beid.connectedServer === null) {
            resolve();
            return;
        }
        beid.connectedServer.getPrimaryService("ae448001-6e87-df33-adb8-51dfd29be725")
            .then(function (service) {
                return service.getCharacteristic("ae448002-6e87-df33-adb8-51dfd29be725");
            })
            .then(function (characteristic) {
                characteristic.removeEventListener('characteristicvaluechanged', beid.handleResponseNotificationsInstance);
                beid.handleResponseNotificationsInstance = null;
                characteristic.stopNotifications().then(function () {
                    resolve();
                });
            });
    });
    return promise;
};

BeIDACR3901U.prototype.close = function (callback, parameter) {
    var beid = this;
    beid.unregisterCardStatusNotificationHandler()
        .then(function () {
            return beid.unregisterBatteryLevelNotificationHandler();
        })
        .then(function () {
            return beid.unregisterResponseNotificationHandler();
        })
        .then(function () {
            if (beid.connectedServer !== null) {
                if (beid.serverDisconnectedHandlerInstance !== null) {
                    beid.connectedServer.device.removeEventListener('gattserverdisconnected', beid.serverDisconnectedHandlerInstance);
                    beid.serverDisconnectedHandlerInstance = null;
                }
                // first set beid.connectedServer to null before disconnecting for Chrome OS delayed removeEventListener.
                var connectedServerInstance = beid.connectedServer;
                beid.connectedServer = null;
                connectedServerInstance.disconnect();
            }
            beid.sessionKey = null;
            beid.longResponseBytesLeft = 0;
            beid.longResponse = null;
            if (typeof callback === "function") {
                callback(parameter);
            }
        });
}

BeIDACR3901U.prototype.applicationAuthentication = function (digestValue, digestAlgo, applicationUrl, signCallback) {
    var beid = this;
    beid.ui.setStatusText("AUTHENTICATING", "INFO");
    // MSE:SET APDU
    beid.sendAPDU([0x00, // cla
        0x22, // ins: Manage Security Environment
        0x41, // p1: Set the signature mode
        0xB6, // p2: value of Digital Signature Template in data field
        0x05, // Lc
        0x04, // length of data structure
        0x80, // algo ref
        0x01, // pkcs#1
        0x84, // tag for private key ref
        0x82 // auth key id
    ], function (mseSetResponse) {
        var mseSetSW = mseSetResponse.subarray(mseSetResponse.length - 2);
        if (mseSetSW[0] !== 0x90 || mseSetSW[1] !== 0x00) {
            console.error("MSE:SET status word: " + beid.Uint8ArrayToHex(mseSetSW));
            beid.ui.setStatusText("ERROR", "ERROR");
            beid.close();
            beid.errorCallback(new Error("status word error"));
            beid.ui.reset();
            return;
        }
        console.log("digest value: " + digestValue);
        var digestInfo = beid.createDigestInfo(digestValue, digestAlgo);
        var cdsApdu = [0x00, // cla
            0x2A, // ins: Perform Security Operation
            0x9E, // p1: PSO:CDS Compute Digital Signature
            0x9A, // p2
            digestInfo.length // Lc
        ].concat(digestInfo);
        cdsApdu.push(0x00); // Le
        beid._cds(cdsApdu, applicationUrl, function (signature) {
            console.log("signature size: " + (signature.length * 8) + " bit");
            signCallback(signature);
        });
    });
};

BeIDACR3901U.prototype.writeData = function (data) {
    var beid = this;
    beid.getServer()
        .then(function (server) {
            server.getPrimaryService("ae448001-6e87-df33-adb8-51dfd29be725")
                .then(function (service) {
                    return service.getCharacteristic("ae448003-6e87-df33-adb8-51dfd29be725");
                })
                .then(function (characteristic) {
                    var queue = Promise.resolve();
                    for (var idx = 0; idx < data.byteLength; idx += 20) {
                        var end = idx + 20;
                        if (end > data.byteLength) {
                            end = data.byteLength;
                        }
                        console.log("idx " + idx + "; end " + end);
                        var writeFunction = function (idx, end) {
                            console.log("function idx " + idx + "; " + end);
                            queue = queue.then(
                                function () {
                                    console.log("write " + idx + " to " + end);
                                    return characteristic.writeValue(data.subarray(idx, end).slice().buffer);
                                }).catch(function (error) {
                                    console.error("error occurred: " + error);
                                    // we have to force a reconnection here when the user retries.
                                    beid.connectedServer = null;
                                    beid.sessionKey = null;
                                    beid.ui.setStatusText(error, "ERROR");
                                    beid.close();
                                    beid.errorCallback(new Error(error));
                                });
                        };
                        writeFunction(idx, end);
                    }
                })
                .catch(function (error) {
                    console.error("error occurred: " + error);
                    // we have to force a reconnection here when the user retries.
                    beid.connectedServer = null;
                    beid.sessionKey = null;
                    beid.ui.setStatusText(error, "ERROR");
                    beid.close();
                    beid.errorCallback(new Error(error));
                });
        });
};

BeIDACR3901U.prototype.dataRequest = function (data) {
    var beid = this;

    beid.progressStatus = !beid.progressStatus;
    beid.ui.progress(beid.progressStatus);

    console.log("data: " + beid.Uint8ArrayToHex(data));
    var paddedData = beid.padToBlocksize(data);
    console.log("padded data: " + beid.Uint8ArrayToHex(paddedData));
    console.log("size padded data: " + paddedData.byteLength);
    var iv = [0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00];
    beid.getSessionKey().then(function (sessionKey) {
        console.log("session key: " + beid.Uint8ArrayToHex(sessionKey));
        var cipher = forge.cipher.createCipher("AES-CBC", Array.from(sessionKey));
        cipher.mode.pad = false;
        cipher.start({ iv: iv });
        cipher.update(new forge.util.ByteBuffer(new Uint8Array(paddedData.slice().buffer)));
        cipher.finish();
        var encryptedData = forge.util.binary.raw.decode(cipher.output.getBytes());
        var dataReq = new Uint8Array(1 + 2 + encryptedData.byteLength + 1);
        dataReq[0] = 0x72;
        dataReq[1] = encryptedData.byteLength + 1;
        dataReq[2] = 0;
        dataReq.set(encryptedData, 3);
        beid.setChecksum(dataReq);

        beid.writeData(dataReq);
    });
};

BeIDACR3901U.prototype.calculateChecksum = function (value) {
    var xor = 0;
    // last byte is the checksum itself
    for (var idx = 0; idx < value.length - 1; idx++) {
        xor ^= value[idx];
    }
    return xor;
};

BeIDACR3901U.prototype.setChecksum = function (value) {
    var beid = this;
    var checksum = beid.calculateChecksum(value);
    value[value.length - 1] = checksum;
};

BeIDACR3901U.prototype.sendAPDU = function (apdu, callback) {
    var beid = this;
    console.log("send APDU...");

    var data = new Uint8Array(1 + 2 + apdu.length + 1);
    data[0] = 0x6f;
    data[1] = apdu.length + 1;
    data[2] = 0;
    data.set(apdu, 3);
    beid.setChecksum(data);

    beid.apduCallback = callback;
    beid.dataRequest(data);
};

BeIDACR3901U.prototype.readBinary = function (offset, length, callback, data) {
    var beid = this;
    var readBinaryApdu = [
        0x00, // CLA
        0xb0, // INS
        (offset >> 8) & 0xff, // P1 = offset high
        (offset >> 0) & 0xff, // P2 = offset low
        length // Le
    ];
    beid.sendAPDU(readBinaryApdu, function (readBinaryApduResponse) {
        var sw = readBinaryApduResponse.subarray(readBinaryApduResponse.length - 2);
        console.log("status word: " + beid.Uint8ArrayToHex(sw));
        if (sw[0] === 0x6b && sw[1] === 0x00) {
            beid.ui.setStatusText("DONE", "SUCCESS");
            callback(data);
            return;
        } else if (sw[0] === 0x6c) {
            length = sw[1];
        } else if (sw[0] !== 0x90 || sw[1] !== 0x00) {
            console.error("status word failure: %s", beid.Uint8ArrayToHex(sw));
            beid.ui.setStatusText("ERROR", "ERROR");
            beid.close();
            beid.errorCallback(new Error("status word failed"));
            beid.ui.reset();
            return;
        }

        if (typeof data !== "undefined") {
            var newData = new Uint8Array(data.length + readBinaryApduResponse.length - 2);
            newData.set(data, 0);
            newData.set(readBinaryApduResponse.subarray(0, readBinaryApduResponse.length - 2), data.length);
            data = newData;
        } else {
            data = readBinaryApduResponse.subarray(0, readBinaryApduResponse.length - 2);
        }
        offset += readBinaryApduResponse.length - 2;
        beid.readBinary(offset, length, callback, data);
    });
};

BeIDACR3901U.prototype.readFile = function (dfName, callback) {
    var beid = this;
    beid.ui.setStatusText("READING", "INFO");
    var selectApdu = [0x00, // cla
        0xa4, // ins
        0x08, // p1 = direct selection by DF name
        0x0c, // p2
        dfName.length // Lc = length of address indication
    ].concat(dfName);
    beid.sendAPDU(selectApdu, function (apduResponse) {
        console.log("select APDU response: " + beid.Uint8ArrayToHex(apduResponse));
        beid.readBinary(0, 0, callback);
    });
};

/**
 * Read the eID identity file.
 */
BeIDACR3901U.prototype.readIdentityFile = function (identityFileCallback) {
    var beid = this;
    var identityDFName = [0x3F, 0x00, 0xDF, 0x01, 0x40, 0x31];
    beid.readFile(identityDFName, function (file) {
        identityFileCallback(file);
    });
};

BeIDACR3901U.prototype.readAddressFile = function (addressFileCallback) {
    var beid = this;
    var addressDFName = [0x3F, 0x00, 0xDF, 0x01, 0x40, 0x33];
    beid.readFile(addressDFName, function (file) {
        addressFileCallback(file);
    });
};

BeIDACR3901U.prototype.readPhotoFile = function (photoFileCallback) {
    var beid = this;
    var photoDFName = [0x3F, 0x00, 0xDF, 0x01, 0x40, 0x35];
    beid.readFile(photoDFName, function (file) {
        photoFileCallback(file);
    });
};

BeIDACR3901U.prototype.readRootCertificateFile = function (certificateFileCallback) {
    var beid = this;
    var rootCertDFName = [0x3F, 0x00, 0xDF, 0x00, 0x50, 0x3B];
    beid.readFile(rootCertDFName, function (file) {
        certificateFileCallback(file);
    });
};

BeIDACR3901U.prototype.readCACertificateFile = function (certificateFileCallback) {
    var beid = this;
    var caCertDFName = [0x3F, 0x00, 0xDF, 0x00, 0x50, 0x3A];
    beid.readFile(caCertDFName, function (file) {
        certificateFileCallback(file);
    });
};

BeIDACR3901U.prototype.readSigningCertificateFile = function (certificateFileCallback) {
    var beid = this;
    var signCertDFName = [0x3F, 0x00, 0xDF, 0x00, 0x50, 0x39];
    beid.readFile(signCertDFName, function (file) {
        certificateFileCallback(file);
    });
};

BeIDACR3901U.prototype.readAuthenticationCertificateFile = function (certificateFileCallback) {
    var beid = this;
    var authCertDFName = [0x3F, 0x00, 0xDF, 0x00, 0x50, 0x38];
    beid.readFile(authCertDFName, function (file) {
        certificateFileCallback(file);
    });
};

BeIDACR3901U.prototype.readNationalRegisterCertificateFile = function (certificateFileCallback) {
    var beid = this;
    var nrCertDFName = [0x3F, 0x00, 0xDF, 0x00, 0x50, 0x3C];
    beid.readFile(nrCertDFName, function (file) {
        certificateFileCallback(file);
    });
};

BeIDACR3901U.prototype.readIdentitySignatureFile = function (identitySignFileCallback) {
    var beid = this;
    beid.readFile([0x3F, 0x00, 0xDF, 0x01, 0x40, 0x32], identitySignFileCallback);
};

BeIDACR3901U.prototype.readAddressSignatureFile = function (addressSignatureFileCallback) {
    var beid = this;
    beid.readFile([0x3F, 0x00, 0xDF, 0x01, 0x40, 0x34], addressSignatureFileCallback);
};

BeIDACR3901U.prototype.logoff = function (callback) {
    var beid = this;
    console.log("sending logoff APDU");
    beid.sendAPDU([0x80, 0xE6, 0x00, 0x00], function (logoffResponse) {
        var logoffSW = logoffResponse.subarray(logoffResponse.length - 2);
        if (logoffSW[0] !== 0x90 || logoffSW[1] !== 0x00) {
            console.error("logoff status word: " + beid.Uint8ArrayToHex(logoffSW));
            beid.ui.setStatusText("ERROR", "ERROR");
            beid.close();
            beid.errorCallback(new Error("status word error"));
            beid.ui.reset();
            return;
        }
        callback();
    });
};

/*
 * Next section contains the tasks scheduler logic
 */

BeIDACR3901U.prototype.createTask = function (taskMethod, fieldName) {
    var beid = this;
    return function (tasks, taskIdx, data, taskCallback) {
        taskMethod.call(beid, function (fieldData) {
            if (typeof fieldName !== "undefined") {
                data[fieldName] = fieldData;
            }

            taskIdx++;
            if (taskIdx > tasks.length - 1) {
                taskCallback(data);
            } else {
                var task = tasks[taskIdx];
                task(tasks, taskIdx, data, taskCallback);
            }
        });
    };
};

BeIDACR3901U.prototype.createAuthenticationTask = function (challenge, applicationUrl, nonceFieldName, signatureFieldName) {
    var beid = this;
    return function (tasks, taskIdx, data, taskCallback) {
        beid.authenticate(challenge, applicationUrl, function (nonce, authenticationSignature) {
            data[nonceFieldName] = nonce;
            data[signatureFieldName] = authenticationSignature;

            taskIdx++;
            if (taskIdx > tasks.length - 1) {
                taskCallback(data);
            } else {
                var task = tasks[taskIdx];
                task(tasks, taskIdx, data, taskCallback);
            }
        });
    };
};

BeIDACR3901U.prototype.createSigningTask = function (digestValue, digestAlgo, applicationUrl, signatureValueFieldName) {
    var beid = this;
    return function (tasks, taskIdx, data, taskCallback) {
        beid.sign(digestValue, digestAlgo, applicationUrl, function (signatureValue) {
            data[signatureValueFieldName] = signatureValue;

            taskIdx++;
            if (taskIdx > tasks.length - 1) {
                taskCallback(data);
            } else {
                var task = tasks[taskIdx];
                task(tasks, taskIdx, data, taskCallback);
            }
        });
    };
};

BeIDACR3901U.prototype.createApplicationAuthenticationTask = function (digestValue, digestAlgo, applicationUrl, signatureValueFieldName) {
    var beid = this;
    return function (tasks, taskIdx, data, taskCallback) {
        beid.applicationAuthentication(digestValue, digestAlgo, applicationUrl, function (signatureValue) {
            data[signatureValueFieldName] = signatureValue;

            taskIdx++;
            if (taskIdx > tasks.length - 1) {
                taskCallback(data);
            } else {
                var task = tasks[taskIdx];
                task(tasks, taskIdx, data, taskCallback);
            }
        });
    };
};

BeIDACR3901U.prototype.createLogoffTask = function () {
    var beid = this;
    return beid.createTask(BeIDACR3901U.prototype.logoff);
};

BeIDACR3901U.prototype.createCloseTask = function () {
    var beid = this;
    return beid.createTask(BeIDACR3901U.prototype.close);
};

BeIDACR3901U.prototype.readAddressFileTask = function (fieldName) {
    var beid = this;
    return beid.createTask(BeIDACR3901U.prototype.readAddressFile, fieldName);
};

BeIDACR3901U.prototype.readAddressSignatureFileTask = function (fieldName) {
    var beid = this;
    return beid.createTask(BeIDACR3901U.prototype.readAddressSignatureFile, fieldName);
};

BeIDACR3901U.prototype.readAuthenticationCertificateFileTask = function (fieldName) {
    var beid = this;
    return beid.createTask(BeIDACR3901U.prototype.readAuthenticationCertificateFile, fieldName);
};

BeIDACR3901U.prototype.readCACertificateFileTask = function (fieldName) {
    var beid = this;
    return beid.createTask(BeIDACR3901U.prototype.readCACertificateFile, fieldName);
};

BeIDACR3901U.prototype.readIdentityFileTask = function (fieldName) {
    var beid = this;
    return beid.createTask(BeIDACR3901U.prototype.readIdentityFile, fieldName);
};

BeIDACR3901U.prototype.readIdentitySignatureFileTask = function (fieldName) {
    var beid = this;
    return beid.createTask(BeIDACR3901U.prototype.readIdentitySignatureFile, fieldName);
};

BeIDACR3901U.prototype.readNationalRegisterCertificateFileTask = function (fieldName) {
    var beid = this;
    return beid.createTask(BeIDACR3901U.prototype.readNationalRegisterCertificateFile, fieldName);
};

BeIDACR3901U.prototype.readPhotoFileTask = function (fieldName) {
    var beid = this;
    return beid.createTask(BeIDACR3901U.prototype.readPhotoFile, fieldName);
};

BeIDACR3901U.prototype.readRootCertificateFileTask = function (fieldName) {
    var beid = this;
    return beid.createTask(BeIDACR3901U.prototype.readRootCertificateFile, fieldName);
};

BeIDACR3901U.prototype.readSigningCertificateFileTask = function (fieldName) {
    var beid = this;
    return beid.createTask(BeIDACR3901U.prototype.readSigningCertificateFile, fieldName);
};

BeIDACR3901U.prototype.runTasks = function (tasks, taskCallback) {
    var data = {};
    if (tasks.length === 0) {
        taskCallback(data);
        return;
    }
    var task = tasks[0];
    task(tasks, 0, data, taskCallback);
};