/*
 * eID Chrome OS project.
 *
 * Copyright 2015-2017 e-Contract.be BVBA. All rights reserved.
 * e-Contract.be BVBA proprietary/confidential. Use is subject to license terms.
 */

/*
 * USB
 * Device Class: Smart Card
 * CCID
 * Specification for Integrated Circuit(s) Cards Interface Devices
 *
 * Revision 1.1
 * April 22rd, 2005
 *
 * http://www.usb.org/developers/docs/devclass_docs/DWG_Smart-Card_CCID_Rev110.pdf
 */

/*
 * ISO/IEC 7816-3:2006(E)
 *
 * Identification cards
 * Integrated circuit cards
 * Part 3: Cards with contacts — Electrical interface and transmission protocols
 *
 * Third edition
 * 2006-11-01
 *
 * http://read.pudn.com/downloads132/doc/comm/563504/ISO-IEC%207816/ISO%2BIEC%207816-3-2006.pdf
 */

"use strict";

BeID.prototype.Uint8ArrayToHex = Uint8ArrayToHex;

function Uint8ArrayToHex(a) {
    var ret = "";
    for (var idx = 0; idx < a.length; idx++) {
        ret += (a[idx] < 16 ? "0" : "") + a[idx].toString(16);
    }
    return ret;
}

var ATR_PATTERN = [0x3b, 0x98, 0x00, 0x40, 0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0xad, 0x13, 0x10];

var ATR_MASK = [0xff, 0xff, 0x00, 0xff, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xf0];

function isBeID(atr) {
    if (atr.length !== ATR_PATTERN.length) {
        return false;
    }
    for (var idx = 0; idx < ATR_PATTERN.length; idx++) {
        if ((atr[idx] & ATR_MASK[idx]) !== ATR_PATTERN[idx]) {
            return false;
        }
    }
    return true;
}

BeID.prototype.sendAPDU = function (apdu, callback) {
    var beid = this;
    var xfrBlock = [
        0x6f, // bMessageType
        (apdu.length >> 0) & 0xff,
        (apdu.length >> 8) & 0xff,
        (apdu.length >> 16) & 0xff,
        (apdu.length >> 24) & 0xff, // dwLength (little-endian)
        0, // bSlot
        0, // bSeq
        0, // bBWI
        0, 0 // wLevelParameter: 0 = short APDU
    ];
    var xfrBlockArray = new Uint8Array(xfrBlock.length + apdu.length);
    xfrBlockArray.set(xfrBlock, 0);
    xfrBlockArray.set(apdu, xfrBlock.length);

    // do not log in production release here due to VERIFY APDU containing the PIN.
    //console.log("APDU REQUEST: (%i) %s", apdu.length, beid.Uint8ArrayToHex(apdu));

    chrome.usb.bulkTransfer(beid.connectionHandle, {
        direction: "out",
        endpoint: beid.bulkOutEndpoint.address,
        data: xfrBlockArray.buffer,
        timeout: 2000
    }, function (transferResultInfo) {
        if (typeof chrome.runtime.lastError !== "undefined" || transferResultInfo.resultCode) {
            console.error("last error: " + chrome.runtime.lastError);
            console.error("APDU out result code: " + transferResultInfo.resultCode);

            beid.ui.setStatusText("ERROR", "ERROR");
            beid.close(beid.errorCallback, new Error("send APDU error: " + chrome.runtime.lastError.message + "; APDU result code: " + transferResultInfo.resultCode));
            beid.ui.reset();
            return;
        }

        //var resultData = new Uint8Array(transferResultInfo.data);
        //console.log("data length: " + resultData.length);
        //console.log("RESPONSE: " + beid.Uint8ArrayToHex(resultData));

        beid.getDataBlock(callback);
    });
};

BeID.prototype.getDataBlock = function (callback, errorCallback) {
    var beid = this;
    console.log("RDR_to_PC_DataBlock");
    // RDR_to_PC_DataBlock
    // keep progress(true) here, then we have a nice blinking eID card icon
    beid.ui.progress(true);
    chrome.usb.bulkTransfer(beid.connectionHandle, {
        direction: "in",
        endpoint: beid.bulkInEndpoint.address,
        timeout: 3000, // 1000 is too short on Chrome OS devices
        length: 512
    }, function (transferResultInfo) {
        if (typeof chrome.runtime.lastError !== "undefined" || transferResultInfo.resultCode) {
            console.error("last error: " + chrome.runtime.lastError);
            console.error("APDU out result code: " + transferResultInfo.resultCode);

            beid.ui.progress(false);
            beid.ui.setStatusText("ERROR", "ERROR");
            beid.close(beid.errorCallback, new Error("error: " + chrome.runtime.lastError.message + "; APDU result code: " + transferResultInfo.resultCode));
            beid.ui.reset();
            return;
        }

        beid.ui.progress(false);
        var resultData = new Uint8Array(transferResultInfo.data);

        if ((resultData[7] & 0x43) !== 0x00) {
            if (typeof errorCallback === "function") {
                errorCallback(resultData);
                return;
            } else {
                // "ICC inactive" or "ICC not present" or "failed"
                console.error("command failed or card in error status, releasing USB. Status=0x%s Error=0x%s", resultData[7].toString(16), resultData[8].toString(16));
                if ((resultData[7] & 0x01) === 0x01) {
                    console.error("inactive ICC");
                }
                if ((resultData[7] & 0x40) === 0x40) { // 01 0000 00
                    console.error("command failed");
                    console.error("error code: 0x" + resultData[8].toString(16));
                    switch (resultData[8]) {
                        case 0xff:
                            console.error("CMD_ABORTED");
                            break;
                        case 0xfe:
                            console.error("ICC_MUTE");
                            break;
                        case 0xfd:
                            console.error("XFR_PARITY_ERROR");
                            break;
                        case 0xfc:
                            console.error("XFR_OVERRUN");
                            break;
                        case 0xfb:
                            console.error("HW_ERROR");
                            break;
                        case 0xf8:
                            console.error("BAD_ATR_TS");
                            break;
                        case 0xf7:
                            console.error("BAD_ATR_TCK");
                            break;
                        case 0xf6:
                            console.error("ICC_PROTOCOL_NOT_SUPPORTED");
                            break;
                        case 0xf5:
                            console.error("ICC_CLASS_NOT_SUPPORTED");
                            break;
                        case 0xf4:
                            console.error("PROCEDURE_BYTE_CONFLICT");
                            break;
                        case 0xf3:
                            console.error("DEACTIVATED_PROTOCOL");
                            break;
                        case 0xf2:
                            console.error("BUSY_WITH_AUTO_SEQUENCE");
                            break;
                        case 0xf0:
                            console.error("PIN_TIMEOUT");
                            break;
                        case 0xef:
                            console.error("PIN_CANCELLED");
                            break;
                        case 0xe0:
                            console.error("CMD_SLOT_BUSY");
                            break;
                    }
                }
                console.log("RESPONSE: (" + resultData.length + ") " + beid.Uint8ArrayToHex(resultData));
                // command failed. At least release the USB interface.
                beid.ui.setStatusText("ERROR", "ERROR");
                beid.close(beid.errorCallback, new Error("command failed: " + beid.Uint8ArrayToHex(resultData)));
                beid.ui.reset();
                return;
            }
        }

        if ((resultData[7] & 0x80) !== 0x00) { // 10 0000 00
            console.log("Time Extension requested");
            // Time Extension is requested
            // wait a bit, and try again
            setTimeout(function () {
                beid.getDataBlock(callback, errorCallback);
            }, 500);
            return;
        }

        var apduRsp = resultData.subarray(10);
        console.log("APDU RESPONSE: (" + apduRsp.length + ") " + beid.Uint8ArrayToHex(apduRsp));
        callback(resultData.subarray(10));
    });
};

BeID.prototype.setParams = function (params, callback) {
    var beid = this;
    // PC_to_RDR_SetParameters
    var setParametersCommand = [
        0x61, // bMessageType
        (params.length >> 0) & 0xff,
        (params.length >> 8) & 0xff,
        (params.length >> 16) & 0xff,
        (params.length >> 24) & 0xff, // dwLength (little-endian)
        0, // bSlot
        0, // bSeq
        0, // bBWI
        0, 0 // wLevelParameter: 0 = short APDU
    ];
    var setParametersArray = new Uint8Array(setParametersCommand.length + params.length);
    setParametersArray.set(setParametersCommand, 0);
    setParametersArray.set(params, setParametersCommand.length);

    chrome.usb.bulkTransfer(beid.connectionHandle, {
        direction: "out",
        endpoint: beid.bulkOutEndpoint.address,
        data: setParametersArray.buffer,
        timeout: 2000
    }, function (transferResultInfo) {
        beid.ui.progress(true);
        if (typeof chrome.runtime.lastError !== "undefined" || transferResultInfo.resultCode) {
            console.error("last error: " + chrome.runtime.lastError);
            console.error("APDU out result code: " + transferResultInfo.resultCode);

            beid.ui.progress(false);
            beid.ui.setStatusText("ERROR", "ERROR");
            beid.close(beid.errorCallback, new Error("error: " + chrome.runtime.lastError));
            beid.ui.reset();
            return;
        }

        chrome.usb.bulkTransfer(beid.connectionHandle, {
            direction: "in",
            endpoint: beid.bulkInEndpoint.address,
            timeout: 2000,
            length: 512
        }, function (transferResultInfo) {
            if (typeof chrome.runtime.lastError !== "undefined" || transferResultInfo.resultCode) {
                console.error("last error: " + chrome.runtime.lastError);
                console.error("APDU out result code: " + transferResultInfo.resultCode);
                beid.ui.progress(false);
                beid.ui.setStatusText("ERROR", "ERROR");
                beid.close(beid.errorCallback, new Error("error: " + chrome.runtime.lastError));
                beid.ui.reset();
                return;
            }

            beid.ui.progress(false);
            var resultData = new Uint8Array(transferResultInfo.data);
            if ((resultData[7] & 0xC3) !== 0x00) {
                console.error("command failed or card in error status, releasing USB. Status=%i Error=%i", resultData[7], resultData[8]);
                // command failed. At least release the USB interface.
                beid.ui.setStatusText("ERROR", "ERROR");
                beid.close(beid.errorCallback, new Error("command failed"));
                beid.ui.reset();
                return;
            }
            callback();
        });
    });
};

BeID.prototype.getResponse = function (dataSize, callback) {
    var beid = this;
    var getResponseApdu = [
        0x00, // CLA
        0xC0, // INS
        0x00, // P1
        0x00, // P2
        dataSize // Le
    ];
    console.log("GET RESPONSE");
    beid.sendAPDU(getResponseApdu, function (getResponseApduResponse) {
        var sw = getResponseApduResponse.subarray(getResponseApduResponse.length - 2);
        if (sw[0] !== 0x90 || sw[1] !== 0x00) {
            console.error("status word failure: " + beid.Uint8ArrayToHex(sw));
            beid.ui.setStatusText("ERROR", "ERROR");
            beid.close(beid.errorCallback, new Error("status word failed"));
            beid.ui.reset();
            return;
        }
        var data = getResponseApduResponse.subarray(0, getResponseApduResponse.length - 2);
        callback(data);
    });
};

BeID.prototype.readBinary = function (offset, length, callback, data) {
    var beid = this;
    var readBinaryApdu = [
        0x00, // CLA
        0xb0, // INS
        (offset >> 8) & 0xff, // P1 = offset high
        (offset >> 0) & 0xff, // P2 = offset low
        length // Le
    ];
    console.log("READ BINARY APDU");
    beid.sendAPDU(readBinaryApdu, function (readBinaryApduResponse) {
        var sw = readBinaryApduResponse.subarray(readBinaryApduResponse.length - 2);
        //console.log("status word: " + beid.Uint8ArrayToHex(sw));
        if (sw[0] === 0x6b && sw[1] === 0x00) {
            beid.ui.setStatusText("DONE", "SUCCESS");
            callback(data);
            return;
        } else if (sw[0] === 0x6c) {
            length = sw[1];
        } else if (sw[0] !== 0x90 || sw[1] !== 0x00) {
            console.error("status word failure: " + beid.Uint8ArrayToHex(sw));
            beid.ui.setStatusText("ERROR", "ERROR");
            beid.close(beid.errorCallback, new Error("READ BINARY APDU: status word failed: " + beid.Uint8ArrayToHex(sw)));
            beid.ui.reset();
            return;
        }

        if (typeof data !== "undefined") {
            var newData = new Uint8Array(data.length + readBinaryApduResponse.length - 2);
            newData.set(data, 0);
            newData.set(readBinaryApduResponse.subarray(0, readBinaryApduResponse.length - 2), data.length);
            data = newData;
        } else {
            data = readBinaryApduResponse.subarray(0, readBinaryApduResponse.length - 2);
        }
        offset += readBinaryApduResponse.length - 2;
        beid.readBinary(offset, length, callback, data);
    });
};

BeID.prototype.readFile = function (dfName, fileCallback) {
    var beid = this;
    var selectApdu = [0x00, // cla
        0xa4, // ins
        0x08, // p1 = direct selection by DF name
        0x0c, // p2
        dfName.length // Lc = length of address indication
    ].concat(dfName);
    beid.ui.setStatusText("READING", "INFO");
    console.log("SELECT APDU");
    beid.sendAPDU(selectApdu, function (selectApduResponse) {
        var sw = selectApduResponse.subarray(selectApduResponse.length - 2);
        if (sw[0] !== 0x90 || sw[1] !== 0x00) {
            console.error("status word failure: " + beid.Uint8ArrayToHex(sw));

            beid.ui.setStatusText("ERROR", "ERROR");
            beid.close(beid.errorCallback, new Error("status word failed"));
            beid.ui.reset();
            return;
        }

        beid.readBinary(0, 0, fileCallback);
    });
};

BeID.prototype.readIdentityFile = function (identityFileCallback) {
    var beid = this;
    beid.readFile([0x3F, 0x00, 0xDF, 0x01, 0x40, 0x31], identityFileCallback);
};

BeID.prototype.readAddressFile = function (addressFileCallback) {
    var beid = this;
    beid.readFile([0x3F, 0x00, 0xDF, 0x01, 0x40, 0x33], addressFileCallback);
};

BeID.prototype.readNationalRegisterCertificateFile = function (natRegCertFileCallback) {
    var beid = this;
    beid.readFile([0x3F, 0x00, 0xDF, 0x00, 0x50, 0x3C], natRegCertFileCallback);
};

BeID.prototype.readIdentitySignatureFile = function (identitySignFileCallback) {
    var beid = this;
    beid.readFile([0x3F, 0x00, 0xDF, 0x01, 0x40, 0x32], identitySignFileCallback);
};

BeID.prototype.readAddressSignatureFile = function (addressSignatureFileCallback) {
    var beid = this;
    beid.readFile([0x3F, 0x00, 0xDF, 0x01, 0x40, 0x34], addressSignatureFileCallback);
};

BeID.prototype.readPhotoFile = function (photoFileCallback) {
    var beid = this;
    beid.readFile([0x3F, 0x00, 0xDF, 0x01, 0x40, 0x35], photoFileCallback);
};

BeID.prototype.readAuthenticationCertificateFile = function (authCertFileCallback) {
    var beid = this;
    beid.readFile([0x3F, 0x00, 0xDF, 0x00, 0x50, 0x38], authCertFileCallback);
};

BeID.prototype.readSigningCertificateFile = function (signCertFileCallback) {
    var beid = this;
    beid.readFile([0x3F, 0x00, 0xDF, 0x00, 0x50, 0x39], signCertFileCallback);
};

BeID.prototype.readCACertificateFile = function (caCertFileCallback) {
    var beid = this;
    beid.readFile([0x3F, 0x00, 0xDF, 0x00, 0x50, 0x3A], caCertFileCallback);
};

BeID.prototype.readRootCertificateFile = function (rootCertFileCallback) {
    var beid = this;
    beid.readFile([0x3F, 0x00, 0xDF, 0x00, 0x50, 0x3B], rootCertFileCallback);
};

BeID.prototype._verifyPinpadResponse = function (applicationUrl, callback) {
    var beid = this;
    chrome.usb.bulkTransfer(beid.connectionHandle, {
        direction: "in",
        endpoint: beid.bulkInEndpoint.address,
        timeout: 2000,
        length: 512
    }, function (transferResultInfo) {
        if (typeof chrome.runtime.lastError !== "undefined" || transferResultInfo.resultCode) {
            console.error("last error: " + chrome.runtime.lastError);
            console.error("APDU out result code: " + transferResultInfo.resultCode);
            if (2 === transferResultInfo.resultCode) {
                // timeout can happen here on some smart card readers (APG8201)
                beid._verifyPinpadResponse(applicationUrl, callback);
                return;
            } else if (1 === transferResultInfo.resultCode) {
                // timeout can happen here on Android
                beid._verifyPinpadResponse(applicationUrl, callback);
                return;
            } else {
                beid.ui.setStatusText("ERROR", "ERROR");
                beid.close(beid.errorCallback, new Error("error"));
                beid.ui.reset();
                return;
            }
        }

        var resultData = new Uint8Array(transferResultInfo.data);
        console.log("VERIFY APDU (pinpad) RESPONSE: (" + resultData.length + ") " + beid.Uint8ArrayToHex(resultData));
        //check the bmCommandStatus for what it is worth (most of the time it is 0 anyway but contains an error or request more time)
        if ((resultData[7] & 0xC3) === 0x00) {
            var verifySW = resultData.subarray(resultData.length - 2);
            if (verifySW[0] === 0x90 && verifySW[1] === 0x00) {
                beid.ui.hidePinpad();
                callback();
            } else if (verifySW[0] === 0x63) {
                var retriesLeft = verifySW[1] & 0x0f;
                beid._verifyPinpad(applicationUrl, callback, retriesLeft);
            } else if (verifySW[0] === 0x69 && verifySW[1] === 0x83) {
                beid.ui.setStatusText("CARD_BLOCKED", "WARNING");
                beid.close(beid.errorCallback, new Error("card blocked"));
                beid.ui.reset();
            } else if (verifySW[0] === 0x64 && verifySW[1] === 0x01) {
                beid.ui.setStatusText("USER_CANCEL", "WARNING");
                beid.close(beid.errorCallback, new Error("user cancel"));
                beid.ui.reset();
            } else if (verifySW[0] === 0x64 && verifySW[1] === 0x01) {
                beid.ui.setStatusText("PINPAD_TIMEOUT", "WARNING");
                beid.close(beid.errorCallback, new Error("pinpad timeout"));
                beid.ui.reset();
            } else {
                console.error("status word failure: " + beid.Uint8ArrayToHex(verifySW));
                beid.ui.setStatusText("ERROR", "ERROR");
                beid.close(beid.errorCallback, new Error("status word failure"));
                beid.ui.reset();
            }
        } else if ((resultData[7] & 0xC3) === 0x80) {
            //Time Extension is requested
            //TODO::Feature #4975
            console.log("waiting for CCID...");
            setTimeout(function () {
                beid._verifyPinpadResponse(applicationUrl, callback);
            }, 200);
        } else if ((resultData[7] & 0xC3) === 0x40) {
            if (resultData[8] === 0xEF) {
                beid.ui.setStatusText("USER_CANCEL", "WARNING");
            } else if (resultData[8] === 0xF0) {
                beid.ui.setStatusText("PINPAD_TIMEOUT", "WARNING");
            } else {
                console.error("ccid error: " + resultData[8]);
                beid.ui.setStatusText("ERROR", "ERROR");
            }
            beid.close(beid.errorCallback, new Error("error"));
            beid.ui.reset();
            return;
        } else {
            console.error("ccid status " + resultData[7] + " error " + resultData[8]);
            beid.ui.setStatusText("ERROR", "ERROR");
            beid.close(beid.errorCallback, new Error("CCID status error"));
            beid.ui.reset();
            return;
        }
    });
};

BeID.prototype._verifyPinpad = function (applicationUrl, verifyCallback, retriesLeft) {
    var beid = this;
    var abPINApdu = [0x00, // cla
        0x20, // ins
        0x00, // p1
        0x01, // p2,
        0x08, // Lc
        0x20,
        0xff,
        0xff,
        0xff,
        0xff,
        0xff,
        0xff,
        0xff];
    var abPINDataStructure = [
        0x00, // bTimeOut
        0x80 | 0x08 | 0x00 | 0x01, // bmFormatString
        /*
         * bmFormatString. bit 7: 1 = system units are bytes
         *
         * bit 6-3: 1 = PIN position in APDU command after Lc, so just after the
         * 0x20 | pinSize.
         *
         * bit 2: 0 = left justify data
         *
         * bit 1-0: 1 = BCD
         */
        0x47, // bmPINBlockString
        /*
         * bmPINBlockString
         *
         * bit 7-4: 4 = PIN length
         *
         * bit 3-0: 7 = PIN block size (7 times 0xff)
         */
        0x04, // bmPINLengthFormat
        /*
         * bmPINLengthFormat. weird... the values do not make any sense to me.
         *
         * bit 7-5: 0 = RFU
         *
         * bit 4: 0 = system units are bits
         *
         * bit 3-0: 4 = PIN length position in APDU
         */
        12, // wPINMaxExtraDigit: Max PIN Len = 12
        4, // wPINMaxExtraDigit, cont.: Min PIN Len = 4
        /*
         * first byte = maximum PIN size in digit
         *
         * second byte = minimum PIN size in digit.
         */
        0x02, // bEntryValidationCondition: Validate Key Press
        /*
         * 0x02 = validation key pressed. So the user must press the green
         * button on his pinpad.
         */
        0x01, // bNumberMessage
        /*
         * 0x01 = message with index in bMsgIndex
         */
        0x04, // wLangId: US
        0x09, // default sub-language
        0x00, // bMsgIndex
        /*
         * 0x00 = PIN insertion prompt
         */
        0x00, 0x00, 0x00 // bTeoPrologue
        /*
         * bTeoPrologue : only significant for T=1 protocol.
         */
    ].concat(abPINApdu);
    var abData = [
        0x00 // bPINOperation = PIN Verification
    ].concat(abPINDataStructure);
    var secure = [
        0x69, // bMessageType
        (abData.length >> 0) & 0xff,
        (abData.length >> 8) & 0xff,
        (abData.length >> 16) & 0xff,
        (abData.length >> 24) & 0xff, // dwLength (little-endian)
        0x00, // bSlot
        0x00, // bSeq
        0x00, // bBWI
        0x00, 0x00 // wLevelParameter
    ].concat(abData);
    beid.ui.setRetriesLeft(retriesLeft);
    beid.ui.showPinpad(applicationUrl);

    console.log("VERIFY APDU (pinpad)");
    chrome.usb.bulkTransfer(beid.connectionHandle, {
        direction: "out",
        endpoint: beid.bulkOutEndpoint.address,
        data: new Uint8Array(secure).buffer,
        timeout: 2000
    }, function (transferResultInfo) {
        if (typeof chrome.runtime.lastError !== "undefined" || transferResultInfo.resultCode) {
            console.error("last error: " + chrome.runtime.lastError);
            console.error("APDU out result code: " + transferResultInfo.resultCode);

            beid.ui.setStatusText("ERROR", "ERROR");
            beid.close(beid.errorCallback, new Error("error: " + chrome.runtime.lastError));
            beid.ui.reset();
            return;
        }

        beid._verifyPinpadResponse(applicationUrl, verifyCallback);
    });
};

BeID.prototype.verify = function (applicationUrl, verifyCallback) {
    var beid = this;
    if (beid.pinVerificationSupport) {
        beid._verifyPinpad(applicationUrl, verifyCallback);
    } else {
        beid._verifyDirect(applicationUrl, verifyCallback);
    }
};

BeID.prototype._verifyDirect = function (applicationUrl, verifyCallback) {
    var beid = this;
    beid.ui.getPIN(applicationUrl, function (pin) {
        var verifyData = [0x20 | pin.length,
            0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF
        ];
        for (var idx = 0; idx < pin.length; idx += 2) {
            var digit1 = pin[idx];
            var digit2;
            if (idx + 1 < pin.length) {
                digit2 = pin[idx + 1];
            } else {
                digit2 = '0' + 0xf;
            }
            var value = ((digit1 - '0' << 4) + (digit2 - '0'));
            verifyData[idx / 2 + 1] = value;
        }
        // MVP:VERIFY
        var verifyApdu = [0x00, // cla
            0x20, // ins
            0x00, // p1
            0x01, // p2
            verifyData.length
        ].concat(verifyData);
        console.log("VERIFY APDU (direct)");
        beid.sendAPDU(verifyApdu, function (verifyApduResponse) {
            var verifySW = verifyApduResponse.subarray(verifyApduResponse.length - 2);
            if (verifySW[0] === 0x90 && verifySW[1] === 0x00) {
                beid.ui.setRetriesLeft();
                verifyCallback();
            } else if (verifySW[0] === 0x63) {
                var retriesLeft = verifySW[1] & 0x0f;
                beid.ui.setRetriesLeft(retriesLeft);
                beid._verifyDirect(applicationUrl, verifyCallback);
            } else if (verifySW[0] === 0x69 && verifySW[1] === 0x83) {
                beid.ui.setStatusText("CARD_BLOCKED", "WARNING");
                beid.close(beid.errorCallback, new Error("card blocked"));
            } else {
                console.error("Error SW " + beid.Uint8ArrayToHex(verifySW));
                beid.ui.setStatusText("ERROR", "ERROR");
                beid.close(beid.errorCallback, new Error("error"));
            }
        });
    });
};

BeID.prototype._cds = function (cdsApdu, applicationUrl, signatureCallback) {
    var beid = this;
    console.log("PSO:CDS APDU");
    beid.sendAPDU(cdsApdu, function (psoCdsResponse) {
        var psoCdsSW = psoCdsResponse.subarray(psoCdsResponse.length - 2);

        if (psoCdsSW[0] === 0x69 && psoCdsSW[1] === 0x82) {
            console.log("PIN required");
            beid.verify(applicationUrl, function () {
                beid._cds(cdsApdu, applicationUrl, signatureCallback);
            });
        } else if (psoCdsSW[0] === 0x61) {
            // GET RESPONSE required
            var signatureSize = psoCdsSW[1];
            beid.getResponse(signatureSize, function (data) {
                beid.ui.setStatusText("DONE", "SUCCESS");
                signatureCallback(data);
            });
        } else if (psoCdsSW[0] === 0x90 && psoCdsSW[1] === 0x00) {
            // some card readers can directly return the signature
            beid.ui.setStatusText("DONE", "SUCCESS");
            signatureCallback(psoCdsResponse.subarray(0, psoCdsResponse.length - 2));
        } else {
            console.error("Error SW " + beid.Uint8ArrayToHex(psoCdsSW));
            beid.ui.setStatusText("ERROR", "ERROR");
            beid.close(beid.errorCallback, new Error("error"));
        }
    });
};

BeID.prototype.authenticate = function (challenge, applicationUrl, authenticationCallback) {
    var beid = this;
    // MSE:SET APDU
    console.log("MSE:SET APDU");
    beid.ui.setStatusText("AUTHENTICATING", "INFO");
    beid.sendAPDU([0x00, // cla
        0x22, // ins: Manage Security Environment
        0x41, // p1: Set the signature mode
        0xB6, // p2: value of Digital Signature Template in data field
        0x05, // Lc
        0x04, // length of data structure
        0x80, // algo ref
        0x02, // pkcs#1 sha1
        0x84, // tag for private key ref
        0x82 // auth key id
    ], function (mseSetResponse) {
        var mseSetSW = mseSetResponse.subarray(mseSetResponse.length - 2);
        if (mseSetSW[0] !== 0x90 || mseSetSW[1] !== 0x00) {
            console.error("MSE:SET status word: " + beid.Uint8ArrayToHex(mseSetSW));
            beid.ui.setStatusText("ERROR", "ERROR");
            beid.close(beid.errorCallback, new Error("status word error"));
            beid.ui.reset();
            return;
        }

        var nonce = new Uint8Array(20);
        window.crypto.getRandomValues(nonce);

        var encodedApplicationUrl = new Uint8Array(applicationUrl.split("").map(function (c) {
            return c.charCodeAt(0);
        }));
        var tbs = new Uint8Array(20 + 20 + encodedApplicationUrl.length);
        tbs.set(nonce, 0);
        tbs.set(challenge, 20);
        tbs.set(encodedApplicationUrl, 20 + 20);

        window.crypto.subtle.digest({
            name: "SHA-1"
        }, tbs).then(function (hash) {
            var hashView = new DataView(hash);
            console.log("hash length: " + hashView.byteLength);
            // PSO:CDS APDU
            var cdsApdu = [0x00, // cla
                0x2A, // ins: Perform Security Operation
                0x9E, // p1: PSO:CDS Compute Digital Signature
                0x9A, // p2
                hashView.byteLength // Lc
            ];
            for (var idx = 0; idx < hashView.byteLength; idx++) {
                cdsApdu.push(hashView.getUint8(idx));
            }
            cdsApdu.push(0x00); // Le
            beid._cds(cdsApdu, applicationUrl, function (signature) {
                console.log("signature size: " + (signature.length * 8) + " bit");
                authenticationCallback(nonce, signature);
            });
        });
    });
};

BeID.prototype.sign = function (digestValue, digestAlgo, applicationUrl, signCallback) {
    var beid = this;
    beid.ui.setStatusText("SIGNING", "INFO");
    // MSE:SET APDU
    beid.sendAPDU([0x00, // cla
        0x22, // ins: Manage Security Environment
        0x41, // p1: Set the signature mode
        0xB6, // p2: value of Digital Signature Template in data field
        0x05, // Lc
        0x04, // length of data structure
        0x80, // algo ref
        0x01, // pkcs#1
        0x84, // tag for private key ref
        0x83 // non-repudiation key id
    ], function (mseSetResponse) {
        var mseSetSW = mseSetResponse.subarray(mseSetResponse.length - 2);
        if (mseSetSW[0] !== 0x90 || mseSetSW[1] !== 0x00) {
            console.error("MSE:SET status word: " + beid.Uint8ArrayToHex(mseSetSW));
            beid.ui.setStatusText("ERROR", "ERROR");
            beid.close(beid.errorCallback, new Error("status word error"));
            beid.ui.reset();
            return;
        }
        console.log("digest value: " + digestValue);
        var digestInfo = beid.createDigestInfo(digestValue, digestAlgo);
        var cdsApdu = [0x00, // cla
            0x2A, // ins: Perform Security Operation
            0x9E, // p1: PSO:CDS Compute Digital Signature
            0x9A, // p2
            digestInfo.length // Lc
        ].concat(digestInfo);
        cdsApdu.push(0x00); // Le
        // for non-repudiation, we know we need a PIN verify anyway.
        beid.verify(applicationUrl, function () {
            beid._cds(cdsApdu, applicationUrl, function (signature) {
                console.log("signature size: " + (signature.length * 8) + " bit");
                signCallback(signature);
            });
        });
    });
};

BeID.prototype.logoff = function (callback) {
    var beid = this;
    console.log("sending logoff APDU");
    beid.sendAPDU([0x80, 0xE6, 0x00, 0x00], function (logoffResponse) {
        var logoffSW = logoffResponse.subarray(logoffResponse.length - 2);
        if (logoffSW[0] !== 0x90 || logoffSW[1] !== 0x00) {
            console.error("logoff status word: " + beid.Uint8ArrayToHex(logoffSW));
            beid.ui.setStatusText("ERROR", "ERROR");
            beid.close(beid.errorCallback, new Error("status word error"));
            beid.ui.reset();
            return;
        }
        callback();
    });
};

BeID.prototype.applicationAuthentication = function (digestValue, digestAlgo, applicationUrl, signCallback) {
    var beid = this;
    beid.ui.setStatusText("AUTHENTICATING", "INFO");
    // MSE:SET APDU
    beid.sendAPDU([0x00, // cla
        0x22, // ins: Manage Security Environment
        0x41, // p1: Set the signature mode
        0xB6, // p2: value of Digital Signature Template in data field
        0x05, // Lc
        0x04, // length of data structure
        0x80, // algo ref
        0x01, // pkcs#1
        0x84, // tag for private key ref
        0x82 // auth key id
    ], function (mseSetResponse) {
        var mseSetSW = mseSetResponse.subarray(mseSetResponse.length - 2);
        if (mseSetSW[0] !== 0x90 || mseSetSW[1] !== 0x00) {
            console.error("MSE:SET status word: " + beid.Uint8ArrayToHex(mseSetSW));
            beid.ui.setStatusText("ERROR", "ERROR");
            beid.close(beid.errorCallback, new Error("status word error"));
            beid.ui.reset();
            return;
        }
        console.log("digest value: " + digestValue);
        var digestInfo = beid.createDigestInfo(digestValue, digestAlgo);
        var cdsApdu = [0x00, // cla
            0x2A, // ins: Perform Security Operation
            0x9E, // p1: PSO:CDS Compute Digital Signature
            0x9A, // p2
            digestInfo.length // Lc
        ].concat(digestInfo);
        cdsApdu.push(0x00); // Le
        beid._cds(cdsApdu, applicationUrl, function (signature) {
            console.log("signature size: " + (signature.length * 8) + " bit");
            signCallback(signature);
        });
    });
};

BeID.prototype.createDigestInfo = function (digestValue, digestAlgo) {
    var digestInfo;
    switch (digestAlgo) {
        case "SHA1":
        case "SHA-1":
            digestInfo = [0x30, 0x21, 0x30, 0x09, 0x06, 0x05, 0x2b, 0x0e, 0x03,
                0x02, 0x1a, 0x05, 0x00, 0x04, 0x14];
            break;
        case "SHA-224":
            digestInfo = [0x30, 0x2d, 0x30, 0x0d, 0x06, 0x09, 0x60, 0x86,
                0x48, 0x01, 0x65, 0x03, 0x04, 0x02, 0x04, 0x05, 0x00, 0x04, 0x1c];
            break;
        case "SHA256":
        case "SHA-256":
            digestInfo = [0x30, 0x31, 0x30, 0x0d, 0x06, 0x09, 0x60, 0x86,
                0x48, 0x01, 0x65, 0x03, 0x04, 0x02, 0x01, 0x05, 0x00, 0x04, 0x20];
            break;
        case "SHA384":
        case "SHA-384":
            digestInfo = [0x30, 0x41, 0x30, 0x0d, 0x06, 0x09, 0x60, 0x86,
                0x48, 0x01, 0x65, 0x03, 0x04, 0x02, 0x02, 0x05, 0x00, 0x04, 0x30];
            break;
        case "SHA512":
        case "SHA-512":
            digestInfo = [0x30, 0x51, 0x30, 0x0d, 0x06, 0x09, 0x60, 0x86,
                0x48, 0x01, 0x65, 0x03, 0x04, 0x02, 0x03, 0x05, 0x00, 0x04, 0x40];
            break;
        case "RIPEMD-128":
            digestInfo = [0x30, 0x1d, 0x30, 0x09, 0x06, 0x05, 0x2b, 0x24, 0x03,
                0x02, 0x02, 0x05, 0x00, 0x04, 0x10];
            break;
        case "RIPEMD-160":
            digestInfo = [0x30, 0x21, 0x30, 0x09, 0x06, 0x05, 0x2b, 0x24, 0x03,
                0x02, 0x01, 0x05, 0x00, 0x04, 0x14];
            break;
        case "RIPEMD-256":
            digestInfo = [0x30, 0x2d, 0x30, 0x09, 0x06, 0x05, 0x2b, 0x24, 0x03,
                0x02, 0x03, 0x05, 0x00, 0x04, 0x20];
            break;
        case "MD5_SHA1":
            digestInfo = [];
            break;
        default:
            console.log("unknown digest algo: " + digestAlgo);
    }
    for (var idx = 0; idx < digestValue.length; idx++) {
        digestInfo.push(digestValue[idx]);
    }
    return digestInfo;
};

/**
 * Releases the USB connection.
 *
 * @param {function} callback optional callback
 * @param {Error} parameter optional callback parameter
 * @returns {undefined}
 */
BeID.prototype.close = function (callback, parameter) {
    var beid = this;
    chrome.usb.releaseInterface(beid.connectionHandle, beid.interfaceNumber, function () {
        console.log("interface released, lastError=" + chrome.runtime.lastError);
        chrome.usb.closeDevice(beid.connectionHandle, function () {
            console.log("device closed, lastError=" + chrome.runtime.lastError);
            if (typeof callback === "function") {
                callback(parameter);
            }
        });
    });
};

BeID.prototype.powerOffAndClose = function (callback) {
    var beid = this;
    console.log("PC_to_RDR_IccPowerOff");
    var data = new Uint8Array([0x63, // bMessageType
        0x00, 0x00, 0x00, 0x00, // dwLength,
        0x00, // bSlot
        0x00, // bSeq
        0x00, 0x00, 0x00 // abRFU
    ]).buffer;
    chrome.usb.bulkTransfer(beid.connectionHandle, {
        direction: "out",
        endpoint: beid.bulkOutEndpoint.address,
        data: data,
        timeout: 2000
    }, function (transferResultInfo) {
        console.log("transfer result info callback");
        console.log("RDR_to_PC_DataBlock");
        chrome.usb.bulkTransfer(beid.connectionHandle, {
            direction: "in",
            endpoint: beid.bulkInEndpoint.address,
            timeout: 2000,
            length: 255
        }, function (transferResultInfo) {
            console.log("transfer result info callback");
            beid.close(callback);
        });
    });
};

BeID.prototype.cardPresentCallback = function (powerSelect) {
    var beid = this;
    if (typeof powerSelect === "undefined") {
        if (beid.scdcDescriptor.automaticVoltageSelection) {
            powerSelect = 0x00; // Automatic Voltage Selection
        } else {
            if (beid.scdcDescriptor.volt18) {
                powerSelect = 0x03; // 1.8 volts
            } else if (beid.scdcDescriptor.volt30) {
                powerSelect = 0x02; // 3 volts
            } else if (beid.scdcDescriptor.volt50) {
                powerSelect = 0x01; // 5 volts
            } else {
                powerSelect = 0x00; // Automatic Voltage Selection
            }
        }
    }
    // PC_to_RDR_IccPowerOn
    console.log("PC_to_RDR_IccPowerOn: power select " + powerSelect);
    var data = new Uint8Array([0x62, // bMessageType
        0x00, 0x00, 0x00, 0x00, // dwLength,
        0x00, // bSlot
        0x00, // bSeq
        powerSelect & 0xff, // bPowerSelect
        0x00, 0x00 // abRFU
    ]).buffer;
    chrome.usb.bulkTransfer(beid.connectionHandle, {
        direction: "out",
        endpoint: beid.bulkOutEndpoint.address,
        data: data,
        timeout: 2000
    }, function (transferResultInfo) {
        console.log("transfer result info callback");
        if (typeof chrome.runtime.lastError !== "undefined" || transferResultInfo.resultCode) {
            console.error("last error: " + chrome.runtime.lastError);
            console.error("APDU out result code: " + transferResultInfo.resultCode);

            beid.ui.setStatusText("ERROR", "ERROR");
            beid.close(beid.errorCallback, new Error("error: " + chrome.runtime.lastError));
            beid.ui.reset();
            return;
        }

        beid.getDataBlock(function (atr) {
            console.log("ATR: " + Uint8ArrayToHex(atr));
            console.log("is BeID: " + isBeID(atr));
            if (isBeID(atr)) {
                //check if we need to configure outself (APDU should support indicate automatic PPS, but most don't)
                if (!beid.automaticPPS && beid.exchangeLevel === "TPDU") {
                    console.log("Manual PPS");

                    //configure card
                    // ISO/IEC 7816-3:2006 - 9.2 PPS request and response
                    beid.sendAPDU([0xFF, //PPSS: fixed FF to indicate PPS
                        0x10, //PPS0: PSS1 set, protocol T=0
                        atr[2], //PPS1:  TA1 of ATR (TS, T0, TA1, ...)
                        (0xFF ^ 0x10 ^ atr[2]) & 0xff //PCK: XOR of all bytes must give 0x00
                    ], function (ppsResponse) {
                        console.log("TPDU for PPS " + ppsResponse);

                        //configure the reader
                        // Protocol Data Structure for Protocol T=0
                        beid.setParams([
                            atr[2], //bmFindexDindex: ATR::TA1
                            0x00, //bmTCCKST0: ignored
                            0x00, //bGuardTimeT0: default
                            atr[4], //bWaitingIntegerT0:  ATR::TC2
                            0x00 //bClockStop: no clock stop
                        ], function () {
                            beid.beidAvailableCallback();
                        });
                    });
                } else {
                    console.log("automatic PPS");
                    beid.beidAvailableCallback();
                }
            } else {
                beid.close();
            }
        }, function (dataBlock) {
            console.log("error occurred");
            if ((dataBlock[7] & 0x40) === 0x40) {
                console.log("command failed on power select " + powerSelect);
                if (powerSelect > 1) {
                    powerSelect = powerSelect - 1;
                    console.log("trying new power select: " + powerSelect);
                    beid.cardPresentCallback(powerSelect);
                    return;
                } else {
                    console.error("all power select tried");
                    beid.ui.setStatusText("ERROR", "ERROR");
                    beid.close(beid.errorCallback, new Error("error on power on: all power select tried"));
                    beid.ui.reset();
                    return;
                }
            } else {
                beid.ui.setStatusText("ERROR", "ERROR");
                beid.close(beid.errorCallback, new Error("error on power on"));
                beid.ui.reset();
                return;
            }
        });
    });
};

BeID.prototype.waitForCardPresentInterrupt = function () {
    var beid = this;
    console.log("calling chrome.usb.interruptTransfer");
    // RDR_to_PC_NotifySlotChange
    chrome.usb.interruptTransfer(beid.connectionHandle, {
        direction: "in",
        endpoint: beid.interruptEndpoint.address,
        timeout: 2000,
        length: 10
    }, function (transferResultInfo) {
        // important to call chrome.runtime.lastError here to capture timeouts
        if (typeof chrome.runtime.lastError !== "undefined" || transferResultInfo.resultCode) {
            console.log("last error: " + chrome.runtime.lastError);
            console.log("result code: " + transferResultInfo.resultCode);
            // OK to continue here: "Transfer time out." can happen here
            beid.waitForCardPresentInterrupt();
            return;
        }

        var resultData = new Uint8Array(transferResultInfo.data);
        console.log("result size: " + resultData.length);
        console.log("result: " + resultData[1]);
        var slot0Changed = resultData[1] & 0x02;
        var slot0Status = resultData[1] & 0x01;
        if (!slot0Changed) {
            beid.waitForCardPresentInterrupt();
        } else {
            if (slot0Status) {
                // check slot status again, since interrupt might also be triggered by inserting card upside down
                beid.waitForCardPresent();
            } else {
                beid.waitForCardPresentInterrupt();
            }
        }
    });
};

BeID.prototype.waitForCardPresent = function () {
    var beid = this;
    // PC_to_RDR_GetSlotStatus
    var data = new Uint8Array([0x65, // bMessageType
        0x00, 0x00, 0x00, 0x00, // dwLength,
        0x00, // bSlot
        0x00, // bSeq
        0x00, 0x00, 0x00 // abRFU
    ]).buffer;
    chrome.usb.bulkTransfer(beid.connectionHandle, {
        direction: "out",
        endpoint: beid.bulkOutEndpoint.address,
        data: data,
        timeout: 2000
    }, function (transferResultInfo) {
        console.log("PC_to_RDR_GetSlotStatus transfer result info callback");
        if (typeof chrome.runtime.lastError !== "undefined" || transferResultInfo.resultCode) {
            // chrome.runtime.lastError is set on Linux when pcscd runs
            console.error("last error: " + chrome.runtime.lastError);
            console.error("error message: " + chrome.runtime.lastError.message);
            console.error("APDU out result code: " + transferResultInfo.resultCode);

            beid.ui.setStatusText("ERROR", "ERROR");
            beid.close(beid.errorCallback, new Error("error: " + chrome.runtime.lastError.message + "; APDU result code: " + transferResultInfo.resultCode));
            beid.ui.reset();
            return;
        }

        // RDR_to_PC_SlotStatus
        chrome.usb.bulkTransfer(beid.connectionHandle, {
            direction: "in",
            endpoint: beid.bulkInEndpoint.address,
            timeout: 2000,
            length: 10
        }, function (transferResultInfo) {
            console.log("RDR_to_PC_SlotStatus transfer result info callback");
            if (typeof chrome.runtime.lastError !== "undefined" || transferResultInfo.resultCode) {
                console.error("last error: " + chrome.runtime.lastError);
                console.error("APDU out result code: " + transferResultInfo.resultCode);

                beid.ui.setStatusText("ERROR", "ERROR");
                beid.close(beid.errorCallback, new Error("error: " + chrome.runtime.lastError));
                beid.ui.reset();
                return;
            }

            console.log(transferResultInfo.data);
            var resultData = new Uint8Array(transferResultInfo.data);
            console.log("slot status: " + resultData[7].toString(16));
            var cardInserted = ((resultData[7] & 0x03) !== 0x02);
            console.log("card inserted: " + cardInserted);
            if (cardInserted) {
                beid.ui.setStatusText();
                beid.cardPresentCallback();
            } else {
                beid.ui.setStatusText("Please insert an eID card...", "INFO");
                if (typeof beid.interruptEndpoint === "undefined") {
                    // if no interrupt endpoint available, we keep polling
                    setTimeout(function () {
                        console.log("after setTimeout of 1000 milliseconds");
                        beid.waitForCardPresent();
                    }, 1000);
                } else {
                    beid.waitForCardPresentInterrupt();
                }
            }
        });
    });
};

function parseSmartCardDeviceClassDescriptor(descriptor, vendorId, productId) {
    console.log("parse smart card device class descriptor: " + Uint8ArrayToHex(descriptor));
    var result = {};
    if (descriptor[0] !== 0x36) {
        console.log("no smart card device class descriptor");
        return result;
    }
    if (descriptor[1] !== 0x21) {
        console.log("incorrect descriptor type");
        return result;
    }
    if ((descriptor[42] & 0x04) === 0x04) {
        result.exchangeLevel = "ShortAndExtendedAPDU";
    } else if ((descriptor[42] & 0x02) === 0x02) {
        result.exchangeLevel = "ShortAPDU";
    } else if ((descriptor[42] & 0x01) === 0x01) {
        result.exchangeLevel = "TPDU";
    } else {
        result.exchangeLevel = "Character";
    }
    console.log("exchange level: " + result.exchangeLevel);
    result.automaticPPS = ((descriptor[40] & 0xC0) !== 0x00);
    result.pinVerificationSupport = ((descriptor[52] & 0x01) === 0x01);
    console.log("number of slots: " + descriptor[4] + 1);
    result.slots = descriptor[4] + 1;
    console.log("maximum CCID busy slots: " + descriptor[53]);

    result.automaticVoltageSelection = ((descriptor[40] & 0x08) === 0x08);
    console.log("automatic voltage selection: " + result.automaticVoltageSelection);
    result.volt50 = ((descriptor[5] & 0x01) === 0x01);
    result.volt30 = ((descriptor[5] & 0x02) === 0x02);
    result.volt18 = ((descriptor[5] & 0x04) === 0x04);
    console.log("5.0 volt supported: " + result.volt50);
    console.log("3.0 volt supported: " + result.volt30);
    console.log("1.8 volt supported: " + result.volt18);

    if (vendorId === 0x072F && productId === 0x8300) {
        // fix for ACR33
        result.pinVerificationSupport = false;
    }
    if (vendorId === 0x0BF8 && productId === 0x1017) {
        // FUJITSUSMARTKEYB
        result.pinVerificationSupport = false;
    }
    if (vendorId === 0x03F0 && productId === 0x0036) {
        // HP_CCIDSMARTCARDKEYBOARD
        result.pinVerificationSupport = false;
    }
    if (vendorId === 0x03F0 && productId === 0x1024) {
        // HPSMARTCARDKEYBOARD
        result.pinVerificationSupport = false;
    }

    return result;
}

function BeID(connectionHandle, bulkInEndpoint, bulkOutEndpoint, interruptEndpoint, interfaceNumber,
    scdcDescriptor, ui, beidAvailableCallback, errorCallback, vendorId, productId) {
    this.connectionHandle = connectionHandle;
    this.bulkInEndpoint = bulkInEndpoint;
    this.bulkOutEndpoint = bulkOutEndpoint;
    this.interruptEndpoint = interruptEndpoint;
    this.interfaceNumber = interfaceNumber;
    this.pinVerificationSupport = scdcDescriptor.pinVerificationSupport;
    this.exchangeLevel = scdcDescriptor.exchangeLevel;
    this.automaticPPS = scdcDescriptor.automaticPPS;
    this.ui = ui;
    this.beidAvailableCallback = beidAvailableCallback;
    this.errorCallback = errorCallback;
    this.vendorId = vendorId;
    this.productId = productId;
    this.scdcDescriptor = scdcDescriptor;
}

function startBeID(devices, ui, beidAvailableCallback, errorCallback) {
    console.log("number of devices: " + devices.length);
    ui.setStatusText("Starting...");
    if (devices.length === 0) {
        // user did not select a device
        // call startSmartCardUsage again here?
        ui.setStatusText("User did not select a device.", "WARNING");
        return;
    }
    var device = devices[0];
    console.log("user selected device: " + device);
    console.log("product name: " + device.productName);
    console.log("manufacturer name: " + device.manufacturerName);
    chrome.usb.openDevice(device, function (connectionHandle) {
        var lastError = chrome.runtime.lastError;
        if (typeof lastError !== "undefined") {
            console.error("last error: " + chrome.runtime.lastError);
            ui.setStatusText("error opening device", "ERROR");
            ui.reset();
            if (typeof errorCallback === "function") {
                errorCallback(new Error("error opening device"));
            }
            return;
        }
        console.log("device opened");
        chrome.usb.listInterfaces(connectionHandle, function (interfaceDescriptors) {
            console.log("interfaceDescriptors: " + interfaceDescriptors);
            var interfaceDescriptor;
            if (interfaceDescriptors.length > 1) {
                for (var idx = 0; idx < interfaceDescriptors.length; idx++) {
                    if (interfaceDescriptors[idx].interfaceClass === 0x0b) {
                        interfaceDescriptor = interfaceDescriptors[idx];
                        break;
                    }
                }
                if (typeof interfaceDescriptor === "undefined") {
                    interfaceDescriptor = interfaceDescriptors[interfaceDescriptors.length - 1];
                }
            } else {
                interfaceDescriptor = interfaceDescriptors[0];
            }
            var extraData = new Uint8Array(interfaceDescriptor.extra_data);
            var scdcDescriptor = parseSmartCardDeviceClassDescriptor(extraData, device.vendorId, device.productId);
            console.log("scdcDescriptor: " + scdcDescriptor);
            var bulkInEndpoint;
            var bulkOutEndpoint;
            var interruptEndpoint;
            interfaceDescriptor.endpoints.forEach(function (endpoint) {
                console.log("endpoint: " + endpoint);
                if (endpoint.type === "bulk") {
                    if (endpoint.direction === "in") {
                        bulkInEndpoint = endpoint;
                    } else if (endpoint.direction === "out") {
                        bulkOutEndpoint = endpoint;
                    }
                } else if (endpoint.type === "interrupt") {
                    interruptEndpoint = endpoint;
                }
            });
            if (typeof bulkInEndpoint === "undefined") {
                console.log("NO BULK IN");
                chrome.usb.closeDevice(connectionHandle, function () {
                    console.log("device closed");
                    if (typeof errorCallback === "function") {
                        errorCallback(new Error("NO BULK IN"));
                    }
                });
                return;
            }
            if (typeof bulkOutEndpoint === "undefined") {
                console.log("NO BULK OUT");
                chrome.usb.closeDevice(connectionHandle, function () {
                    console.log("device closed");
                    if (typeof errorCallback === "function") {
                        errorCallback(new Error("NO BULK OUT"));
                    }
                });
                return;
            }
            chrome.usb.claimInterface(connectionHandle, interfaceDescriptor.interfaceNumber, function () {
                if (typeof chrome.runtime.lastError !== "undefined") {
                    console.error("last error: " + chrome.runtime.lastError);
                    var lastErrorMessage = chrome.runtime.lastError.message;
                    ui.setStatusText("Error claiming interface", "ERROR");
                    chrome.usb.closeDevice(connectionHandle, function () {
                        console.log("device closed");
                        if (typeof errorCallback === "function") {
                            errorCallback(new Error("Error claiming interface: " + lastErrorMessage));
                        }
                    });
                    return;
                }
                console.log("interface claimed");
                var vendorId = device.vendorId.toString(16);
                var productId = device.productId.toString(16);
                var beid = new BeID(connectionHandle, bulkInEndpoint, bulkOutEndpoint, interruptEndpoint, interfaceDescriptor.interfaceNumber,
                    scdcDescriptor, ui, beidAvailableCallback, errorCallback, vendorId, productId);
                beid.waitForCardPresent();
            });
        });
    });
}

/*
 * Next section contains the tasks scheduler logic
 */

BeID.prototype.createTask = function (taskMethod, fieldName) {
    var beid = this;
    return function (tasks, taskIdx, data, taskCallback) {
        taskMethod.call(beid, function (fieldData) {
            if (typeof fieldName !== "undefined") {
                data[fieldName] = fieldData;
            }

            taskIdx++;
            if (taskIdx > tasks.length - 1) {
                taskCallback(data);
            } else {
                var task = tasks[taskIdx];
                task(tasks, taskIdx, data, taskCallback);
            }
        });
    };
};

BeID.prototype.createAuthenticationTask = function (challenge, applicationUrl, nonceFieldName, signatureFieldName) {
    var beid = this;
    return function (tasks, taskIdx, data, taskCallback) {
        beid.authenticate(challenge, applicationUrl, function (nonce, authenticationSignature) {
            data[nonceFieldName] = nonce;
            data[signatureFieldName] = authenticationSignature;

            taskIdx++;
            if (taskIdx > tasks.length - 1) {
                taskCallback(data);
            } else {
                var task = tasks[taskIdx];
                task(tasks, taskIdx, data, taskCallback);
            }
        });
    };
};

BeID.prototype.createSigningTask = function (digestValue, digestAlgo, applicationUrl, signatureValueFieldName) {
    var beid = this;
    return function (tasks, taskIdx, data, taskCallback) {
        beid.sign(digestValue, digestAlgo, applicationUrl, function (signatureValue) {
            data[signatureValueFieldName] = signatureValue;

            taskIdx++;
            if (taskIdx > tasks.length - 1) {
                taskCallback(data);
            } else {
                var task = tasks[taskIdx];
                task(tasks, taskIdx, data, taskCallback);
            }
        });
    };
};

BeID.prototype.createApplicationAuthenticationTask = function (digestValue, digestAlgo, applicationUrl, signatureValueFieldName) {
    var beid = this;
    return function (tasks, taskIdx, data, taskCallback) {
        beid.applicationAuthentication(digestValue, digestAlgo, applicationUrl, function (signatureValue) {
            data[signatureValueFieldName] = signatureValue;

            taskIdx++;
            if (taskIdx > tasks.length - 1) {
                taskCallback(data);
            } else {
                var task = tasks[taskIdx];
                task(tasks, taskIdx, data, taskCallback);
            }
        });
    };
};

BeID.prototype.createLogoffTask = function () {
    var beid = this;
    return beid.createTask(BeID.prototype.logoff);
};

BeID.prototype.createCloseTask = function () {
    var beid = this;
    return beid.createTask(BeID.prototype.close);
};

BeID.prototype.readAddressFileTask = function (fieldName) {
    var beid = this;
    return beid.createTask(BeID.prototype.readAddressFile, fieldName);
};

BeID.prototype.readAddressSignatureFileTask = function (fieldName) {
    var beid = this;
    return beid.createTask(BeID.prototype.readAddressSignatureFile, fieldName);
};

BeID.prototype.readAuthenticationCertificateFileTask = function (fieldName) {
    var beid = this;
    return beid.createTask(BeID.prototype.readAuthenticationCertificateFile, fieldName);
};

BeID.prototype.readCACertificateFileTask = function (fieldName) {
    var beid = this;
    return beid.createTask(BeID.prototype.readCACertificateFile, fieldName);
};

BeID.prototype.readIdentityFileTask = function (fieldName) {
    var beid = this;
    return beid.createTask(BeID.prototype.readIdentityFile, fieldName);
};

BeID.prototype.readIdentitySignatureFileTask = function (fieldName) {
    var beid = this;
    return beid.createTask(BeID.prototype.readIdentitySignatureFile, fieldName);
};

BeID.prototype.readNationalRegisterCertificateFileTask = function (fieldName) {
    var beid = this;
    return beid.createTask(BeID.prototype.readNationalRegisterCertificateFile, fieldName);
};

BeID.prototype.readPhotoFileTask = function (fieldName) {
    var beid = this;
    return beid.createTask(BeID.prototype.readPhotoFile, fieldName);
};

BeID.prototype.readRootCertificateFileTask = function (fieldName) {
    var beid = this;
    return beid.createTask(BeID.prototype.readRootCertificateFile, fieldName);
};

BeID.prototype.readSigningCertificateFileTask = function (fieldName) {
    var beid = this;
    return beid.createTask(BeID.prototype.readSigningCertificateFile, fieldName);
};

BeID.prototype.runTasks = function (tasks, taskCallback) {
    var data = {};
    if (tasks.length === 0) {
        taskCallback(data);
        return;
    }
    var task = tasks[0];
    task(tasks, 0, data, taskCallback);
};