/*
 * eID Android project.
 *
 * Copyright 2016-2018 e-Contract.be BVBA. All rights reserved.
 * e-Contract.be BVBA proprietary/confidential. Use is subject to license terms.
 */

"use strict";

var app = {
    btbeid: new BeIDACR3901U(),
    initialize: function () {
        document.addEventListener('deviceready', this.onDeviceReady, false);
    },
    onReadyStateChange: function () {
        if (this.readyState !== 4) {
            return;
        }
        if (this.status === 200) {
            app.ui.setStatusText("Processing task...", "INFO");
            console.log("response: " + this.responseText);
            var response = JSON.parse(this.responseText);
            switch (response.type) {
                case "identification-request":
                    var tasks = [
                        app.beid.readIdentityFileTask("identity"),
                        app.beid.readIdentitySignatureFileTask("identitySignature"),
                        app.beid.readNationalRegisterCertificateFileTask("nationalRegisterCertificate"),
                        app.beid.readRootCertificateFileTask("rootCertificate")
                    ];
                    if (response.includeAddress) {
                        tasks.push(app.beid.readAddressFileTask("address"));
                        tasks.push(app.beid.readAddressSignatureFileTask("addressSignature"));
                    }
                    if (response.includeAuthenticationCertificate || response.includeSignatureCertificate) {
                        tasks.push(app.beid.readCACertificateFileTask("caCertificate"));
                    }
                    if (response.includeAuthenticationCertificate) {
                        tasks.push(app.beid.readAuthenticationCertificateFileTask("authenticationCertificate"));
                    }
                    if (response.includeSignatureCertificate) {
                        tasks.push(app.beid.readSigningCertificateFileTask("signatureCertificate"));
                    }
                    if (response.includePhoto) {
                        tasks.push(app.beid.readPhotoFileTask("photo"));
                    }
                    app.beid.runTasks(tasks, function (data) {
                        var request = {
                            type: "identification-response"
                        };
                        for (var prop in data) {
                            request[prop] = btoa(String.fromCharCode.apply(null, data[prop]));
                        }
                        app.sendToServer(request);
                    });
                    break;
                case "authentication-request":
                    var encodedChallenge = response.challenge;
                    var challenge = new Uint8Array(atob(encodedChallenge).split("").map(function (c) {
                        return c.charCodeAt(0);
                    }));
                    var applicationUrl = app.serverName;
                    var preLogoff = response.preLogoff;
                    var postLogoff = response.postLogoff;
                    var tasks = [
                        app.beid.readRootCertificateFileTask("rootCertificate"),
                        app.beid.readAuthenticationCertificateFileTask("authenticationCertificate"),
                        app.beid.readCACertificateFileTask("caCertificate")
                    ];
                    if (preLogoff) {
                        tasks.push(app.beid.createLogoffTask());
                    }
                    tasks.push(app.beid.createAuthenticationTask(challenge, applicationUrl, "nonce", "signature"));
                    if (postLogoff) {
                        tasks.push(app.beid.createLogoffTask());
                    }
                    app.beid.runTasks(tasks, function (data) {
                        var protocolData = {
                            type: "authentication-response",
                            applicationUrl: applicationUrl
                        };
                        for (var prop in data) {
                            protocolData[prop] = btoa(String.fromCharCode.apply(null, data[prop]));
                        }
                        app.sendToServer(protocolData);
                    });
                    break;
                case "signing-request":
                    var encodedDigestValue = response.digestValue;
                    var digestValue = new Uint8Array(atob(encodedDigestValue).split("").map(function (c) {
                        return c.charCodeAt(0);
                    }));
                    var digestAlgo = response.digestAlgo;
                    var applicationUrl = app.serverName;
                    var postLogoff = response.postLogoff;
                    var tasks = [
                        app.beid.readRootCertificateFileTask("rootCertificate"),
                        app.beid.readSigningCertificateFileTask("signingCertificate"),
                        app.beid.readCACertificateFileTask("caCertificate"),
                        app.beid.createSigningTask(digestValue, digestAlgo, applicationUrl, "signatureValue")
                    ];
                    if (postLogoff) {
                        tasks.push(app.beid.createLogoffTask());
                    }
                    app.beid.runTasks(tasks, function (data) {
                        var protocolData = {
                            type: "signing-response",
                            applicationUrl: applicationUrl
                        };
                        for (var prop in data) {
                            protocolData[prop] = btoa(String.fromCharCode.apply(null, data[prop]));
                        }
                        app.sendToServer(protocolData);
                    });
                    break;
                case "application-authentication-request":
                    var encodedDigestValue = response.digestValue;
                    var digestValue = new Uint8Array(atob(encodedDigestValue).split("").map(function (c) {
                        return c.charCodeAt(0);
                    }));
                    var digestAlgo = response.digestAlgo;
                    var applicationName = response.applicationName;
                    var preLogoff = response.preLogoff;
                    var postLogoff = response.postLogoff;
                    var tasks = [];
                    if (preLogoff) {
                        tasks.push(app.beid.createLogoffTask());
                    }
                    tasks.push(app.beid.createApplicationAuthenticationTask(digestValue, digestAlgo, applicationName, "signatureValue"));
                    if (postLogoff) {
                        tasks.push(app.beid.createLogoffTask());
                    }
                    app.beid.runTasks(tasks, function (data) {
                        var protocolData = {
                            type: "application-authentication-response"
                        };
                        for (var prop in data) {
                            protocolData[prop] = btoa(String.fromCharCode.apply(null, data[prop]));
                        }
                        app.sendToServer(protocolData);
                    });
                    break;
                case "stop":
                    console.log("stop");
                    app.beid.powerOffAndClose(function () {
                        navigator.app.exitApp();
                    });
                    break;
                case "error-acknowledge":
                    navigator.app.exitApp();
                    break;
            }
        } else {
            app.ui.setStatusText("Server error: " + this.status, "ERROR");
        }
    },
    sendToServer: function (request) {
        app.ui.setStatusText("Transmitting data...", "INFO");
        if (typeof app.serverUrl === "undefined") {
            app.ui.setStatusText("Missing server URL!", "ERROR");
            return;
        }
        console.log("application url: " + app.serverUrl);
        var client = new XMLHttpRequest();
        client.onreadystatechange = app.onReadyStateChange;
        client.open("POST", app.serverUrl, true);
        client.send(JSON.stringify(request));
    },
    onDeviceReady: function () {
        if ("undefined" === typeof window.crypto.subtle) {
            // patch for older Androids
            window.crypto.subtle = {};
            window.crypto.subtle.digest = function (options, message) {
                return {
                    then: function (callback) {
                        var shaObj = new jsSHA("SHA-1", "BYTES");
                        for (var idx = 0; idx < message.length; idx++) {
                            shaObj.update(String.fromCharCode(message[idx]));
                        }
                        var hash = shaObj.getHash("BYTES");
                        var result = new ArrayBuffer(20);
                        var resultView = new DataView(result);
                        for (var idx = 0; idx < hash.length; idx++) {
                            resultView.setUint8(idx, hash.charCodeAt(idx));
                        }
                        callback(result);
                    }
                };
            };
        }
        window.plugins.intent.setNewIntentHandler(app.handleIntent);
        window.plugins.intent.getCordovaIntent(app.handleIntent);
        document.addEventListener("pause", app.onPause, false);
        document.addEventListener("resume", app.onResume, false);
        document.querySelector("#readCardButton").addEventListener("click", app.readCardClicked);
        document.querySelector("#readCardBluetoothButton").addEventListener("click", app.readCardBluetoothClicked);

        $(document).on("change", "[name='readerSelection']", function () {
            console.log("reader selection: " + $(this).val());
            window.localStorage.setItem("readerSelection", $(this).val());
        });

        var initialReaderSelection = window.localStorage.getItem("readerSelection");
        if (initialReaderSelection === null) {
            initialReaderSelection = DEFAULT_READER_SELECTION;
        }
        console.log("initial reader selection: " + initialReaderSelection);
        $("[name='readerSelection'][value='" + initialReaderSelection + "']").attr("checked", "checked");
        initOffline();
    },
    readCardBluetoothClicked: function () {
        console.log("read card bluetooth clicked");
        document.querySelector("#eID_name").innerHTML = "";
        document.querySelector("#eID_givenName").innerHTML = "";
        document.querySelector("#eID_placeOfBirth").innerHTML = "";
        document.querySelector("#eID_dateOfBirth").innerHTML = "";
        document.querySelector("#eID_nationalNumber").innerHTML = "";
        document.querySelector("#eID_cardNumber").innerHTML = "";
        document.querySelector("#eID_documentType").innerHTML = "";
        document.querySelector("#eID_validity").innerHTML = "";
        document.querySelector("#eID_nationality").innerHTML = "";
        document.querySelector("#eID_gender").innerHTML = "";

        document.querySelector("#eID_streetAndNumber").innerHTML = "";
        document.querySelector("#eID_zip").innerHTML = "";
        document.querySelector("#eID_municipality").innerHTML = "";

        var canvas = document.querySelector("#eID_photo");
        var context = canvas.getContext("2d");
        context.fillStyle = "#ffffff";
        context.fillRect(0, 0, 140, 200);

        document.querySelector("#readCardButton").disabled = true;
        document.querySelector("#readCardBluetoothButton").disabled = true;

        app.ui.setStatusText("Please enable your bluetooth smart card reader...", "INFO");
        app.btbeid.start(app.ui, function () {
            console.log("bluetooth reader available");
            app.btbeid.cardPowerOn(function (atr) {
                app.btbeid.readIdentityFile(function (identityFile) {
                    console.log("identity file done");
                    var identity = parseIdentityFile(identityFile);
                    document.querySelector("#eID_name").innerHTML = identity.name;
                    document.querySelector("#eID_givenName").innerHTML = identity.firstName;
                    document.querySelector("#eID_placeOfBirth").innerHTML = identity.placeOfBirth;
                    var dateOfBirthOptions = {
                        year: 'numeric',
                        month: 'long',
                        day: 'numeric'
                    };
                    document.querySelector("#eID_dateOfBirth").innerHTML = identity.dateOfBirth.toLocaleDateString('nl-BE', dateOfBirthOptions);
                    document.querySelector("#eID_nationalNumber").innerHTML = identity.nationalNumber;
                    document.querySelector("#eID_cardNumber").innerHTML = identity.cardNumber;
                    document.querySelector("#eID_documentType").innerHTML = identity.documentType;
                    var validityOptions = {
                        year: 'numeric',
                        month: 'numeric',
                        day: 'numeric'
                    };
                    var validity = identity.cardValidityDateBegin.toLocaleDateString("nl-BE", validityOptions) + " until " +
                            identity.cardValidityDateEnd.toLocaleDateString("nl-BE", validityOptions);
                    document.querySelector("#eID_validity").innerHTML = validity;
                    document.querySelector("#eID_nationality").innerHTML = identity.nationality;
                    document.querySelector("#eID_gender").innerHTML = identity.gender;
                    app.btbeid.readAddressFile(function (addressFile) {
                        var address = parseAddressFile(addressFile);
                        document.querySelector("#eID_streetAndNumber").innerHTML = address.streetAndNumber;
                        document.querySelector("#eID_zip").innerHTML = address.zip;
                        document.querySelector("#eID_municipality").innerHTML = address.municipality;

                        var canvas = document.querySelector("#eID_photo");
                        var context = canvas.getContext("2d");
                        context.font = "10px Arial";
                        context.fillStyle = "#000000";
                        context.fillText("Loading...", 0, 30);

                        app.btbeid.readPhotoFile(function (photo) {
                            console.log("receiving eID photo");
                            var canvas = document.querySelector("#eID_photo");
                            var image = new window.Image();
                            image.src = "data:image/jpeg;base64," + btoa(String.fromCharCode.apply(null, photo));
                            image.onload = function () {
                                var context = canvas.getContext("2d");
                                context.drawImage(image, 0, 0);
                            };
                            app.btbeid.powerOffAndClose(function () {
                                document.querySelector("#readCardButton").disabled = false;
                                document.querySelector("#readCardBluetoothButton").disabled = false;
                            });
                        });
                    });
                });
            });
        }, function (error) {
            console.error("error");
            console.error("error message: " + error.message);
            console.error("error stack: " + error.stack);
            app.ui.setStatusText(error.message, "ERROR");
            document.querySelector("#readCardButton").disabled = false;
            document.querySelector("#readCardBluetoothButton").disabled = false;
        });
    },
    readCardClicked: function () {
        console.log("read card clicked");
        document.querySelector("#eID_name").innerHTML = "";
        document.querySelector("#eID_givenName").innerHTML = "";
        document.querySelector("#eID_placeOfBirth").innerHTML = "";
        document.querySelector("#eID_dateOfBirth").innerHTML = "";
        document.querySelector("#eID_nationalNumber").innerHTML = "";
        document.querySelector("#eID_cardNumber").innerHTML = "";
        document.querySelector("#eID_documentType").innerHTML = "";
        document.querySelector("#eID_validity").innerHTML = "";
        document.querySelector("#eID_nationality").innerHTML = "";
        document.querySelector("#eID_gender").innerHTML = "";

        document.querySelector("#eID_streetAndNumber").innerHTML = "";
        document.querySelector("#eID_zip").innerHTML = "";
        document.querySelector("#eID_municipality").innerHTML = "";

        var canvas = document.querySelector("#eID_photo");
        var context = canvas.getContext("2d");
        context.fillStyle = "#ffffff";
        context.fillRect(0, 0, 140, 200);

        document.querySelector("#readCardButton").disabled = true;
        document.querySelector("#readCardBluetoothButton").disabled = true;

        selectCordovaAndroidCCIDDevice(function (devices) {
            if (devices.length === 0) {
                console.log("no device available");
                app.ui.setStatusText("Please connect a smart card reader...", "WARNING");
                setTimeout(function () {
                    // next construct avoids burning battery
                    if (!app.pauzed) {
                        app.readCardClicked();
                    }
                }, 500);
                return;
            }
            console.log("start reading...");
            startBeID(devices, app.ui, function () {
                console.log("started beid");
                var beid = this;
                beid.readIdentityFile(function (identityFile) {
                    console.log("identity file done");
                    var identity = parseIdentityFile(identityFile);
                    document.querySelector("#eID_name").innerHTML = identity.name;
                    document.querySelector("#eID_givenName").innerHTML = identity.firstName;
                    document.querySelector("#eID_placeOfBirth").innerHTML = identity.placeOfBirth;
                    var dateOfBirthOptions = {
                        year: 'numeric',
                        month: 'long',
                        day: 'numeric'
                    };
                    document.querySelector("#eID_dateOfBirth").innerHTML = identity.dateOfBirth.toLocaleDateString('nl-BE', dateOfBirthOptions);
                    document.querySelector("#eID_nationalNumber").innerHTML = identity.nationalNumber;
                    document.querySelector("#eID_cardNumber").innerHTML = identity.cardNumber;
                    document.querySelector("#eID_documentType").innerHTML = identity.documentType;
                    var validityOptions = {
                        year: 'numeric',
                        month: 'numeric',
                        day: 'numeric'
                    };
                    var validity = identity.cardValidityDateBegin.toLocaleDateString("nl-BE", validityOptions) + " until " +
                            identity.cardValidityDateEnd.toLocaleDateString("nl-BE", validityOptions);
                    document.querySelector("#eID_validity").innerHTML = validity;
                    document.querySelector("#eID_nationality").innerHTML = identity.nationality;
                    document.querySelector("#eID_gender").innerHTML = identity.gender;
                    beid.readAddressFile(function (addressFile) {
                        var address = parseAddressFile(addressFile);
                        document.querySelector("#eID_streetAndNumber").innerHTML = address.streetAndNumber;
                        document.querySelector("#eID_zip").innerHTML = address.zip;
                        document.querySelector("#eID_municipality").innerHTML = address.municipality;

                        var canvas = document.querySelector("#eID_photo");
                        var context = canvas.getContext("2d");
                        context.font = "10px Arial";
                        context.fillStyle = "#000000";
                        context.fillText("Loading...", 0, 30);

                        beid.readPhotoFile(function (photo) {
                            console.log("receiving eID photo");
                            var canvas = document.querySelector("#eID_photo");
                            var image = new window.Image();
                            image.src = "data:image/jpeg;base64," + btoa(String.fromCharCode.apply(null, photo));
                            image.onload = function () {
                                var context = canvas.getContext("2d");
                                context.drawImage(image, 0, 0);
                            };
                            beid.powerOffAndClose(function () {
                                document.querySelector("#readCardButton").disabled = false;
                                document.querySelector("#readCardBluetoothButton").disabled = false;
                            });
                        });
                    });
                });
            }, function () {
                app.ui.setStatusText("An error occurred.", "ERROR");
                document.querySelector("#readCardButton").disabled = false;
                document.querySelector("#readCardBluetoothButton").disabled = false;
            });
        });
    },
    onPause: function () {
        console.log("pause");
        app.pauzed = true;
    },
    onResume: function () {
        console.log("resume");
        app.pauzed = false;
        document.querySelector("#readCardButton").disabled = false;
        document.querySelector("#readCardBluetoothButton").disabled = false;
        app.ui.setStatusText();
    },
    handleIntent: function (intent) {
        app.ui = new UI();
        if ("android.intent.action.VIEW" === intent.action) {
            var beidUrl = intent.data;
            if (beidUrl === "beid:") {
                // in case of offline activity launch
                document.querySelector("#main").hidden = false;
                return;
            } else {
                document.querySelector("#main").hidden = true;
                var readerSelection = window.localStorage.getItem("readerSelection");
                if (readerSelection === null) {
                    readerSelection = DEFAULT_READER_SELECTION;
                }
                switch (readerSelection) {
                    case "both":
                        console.log("both options");
                        break;
                    case "usb":
                        app.selectDevice(beidUrl);
                        break;
                    case "bluetooth":
                        console.log("bluetooth");
                        app.selectBluetoothDevice(beidUrl);
                        break;
                }
            }
        } else {
            document.querySelector("#main").hidden = false;
        }
    },
    selectDevice: function (beidUrl) {
        console.log("select device: " + beidUrl);
        selectCordovaAndroidCCIDDevice(function (devices) {
            if (devices.length === 0) {
                console.log("no device available");
                app.ui.setStatusText("Please connect a smart card reader...", "WARNING");
                setTimeout(function () {
                    // next construct avoids burning battery
                    if (!app.pauzed) {
                        app.selectDevice(beidUrl);
                    }
                }, 500);
                return;
            }
            app.ui.setStatusText("Starting...", "INFO");
            startBeID(devices, app.ui, function () {
                app.beid = this;
                app.serverUrl = EID_ANDROID_CONFIG.scheme + beidUrl.substring("beid".length);
                app.serverName = app.serverUrl.substring(0, app.serverUrl.indexOf("?"));
                app.sendToServer({
                    type: "start",
                    userAgent: navigator.userAgent,
                    applicationVersion: EID_ANDROID_CONFIG.version,
                    vendorId: app.beid.vendorId,
                    productId: app.beid.productId
                });
            }, function (errorObject) {
                app.ui.setStatusText("An error occurred.", "ERROR");
                // app.serverUrl is required for app.sendToServer
                app.serverUrl = EID_ANDROID_CONFIG.scheme + beidUrl.substring("beid".length);
                app.sendToServer({
                    type: "error",
                    errorMessage: errorObject.message,
                    errorStack: errorObject.stack
                });
            });
        });
    },
    selectBluetoothDevice: function (beidUrl) {
        app.ui.setStatusText("Please enable your bluetooth smart card reader...", "INFO");
        app.btbeid.start(app.ui, function () {
            console.log("bluetooth reader available");
            app.btbeid.cardPowerOn(function (atr) {
                app.beid = app.btbeid;
                app.serverUrl = EID_ANDROID_CONFIG.scheme + beidUrl.substring("beid".length);
                app.serverName = app.serverUrl.substring(0, app.serverUrl.indexOf("?"));
                app.sendToServer({
                    type: "start",
                    userAgent: navigator.userAgent,
                    applicationVersion: EID_ANDROID_CONFIG.version,
                    vendorId: app.beid.vendorId,
                    productId: app.beid.productId
                });
            });
        }, function (errorObject) {
            app.ui.setStatusText(errorObject.message, "ERROR");
            // app.serverUrl is required for app.sendToServer
            app.serverUrl = EID_ANDROID_CONFIG.scheme + beidUrl.substring("beid".length);
            app.sendToServer({
                type: "error",
                errorMessage: errorObject.message,
                errorStack: errorObject.stack
            });
        });
    }
};

app.initialize();

function UI() {
}

UI.prototype.setStatusText = function (status, level) {
    if (typeof status === "undefined") {
        status = "";
    }
    var cssLevel;
    var statusEl = document.querySelector("#status");
    if (typeof level === "undefined") {
        cssLevel = "INFO";
    } else {
        cssLevel = level.toUpperCase();
    }
    statusEl.className = cssLevel;
    statusEl.innerHTML = status;
};

UI.prototype.getPIN = function (applicationUrl, pinCallback) {
    document.querySelector("#pinApplicationUrl").innerHTML = applicationUrl;
    document.querySelector("#pinDialog").style.display = "block";
    var pinClickEventListener = function () {
        document.querySelector("#pinDialog").style.display = "none";
        var pin = document.querySelector("#pin").value;
        document.querySelector("#pin").value = "";
        // important to remove ourselves, else next time, we trigger two callbacks
        document.querySelector("#pinButton").removeEventListener("click", pinClickEventListener);
        document.querySelector("#pin").removeEventListener("keydown", pinKeyPressEventListener);
        pinCallback(pin);
    };
    var pinKeyPressEventListener = function (event) {
        console.log("keypress event");
        if (event.keyCode === 13) {
            pinClickEventListener();
        }
    };
    document.querySelector("#pinButton").addEventListener("click", pinClickEventListener);
    document.querySelector("#pin").addEventListener("keydown", pinKeyPressEventListener);
    document.querySelector("#pin").focus();
    Keyboard.show();
};

UI.prototype.showPinpad = function (applicationUrl) {
    document.querySelector("#pinpadApplicationUrl").innerHTML = applicationUrl;
    document.querySelector("#pinpadDialog").style.display = "block";
};

UI.prototype.hidePinpad = function () {
    document.querySelector("#pinpadDialog").style.display = "none";
};

UI.prototype.reset = function () {
    clearBattery("battery");
};

UI.prototype.progress = function (active) {
    var progressElement = document.querySelector("#progress");
    if (active) {
        progressElement.style.display = "block";
    } else {
        progressElement.style.display = "none";
    }
};

UI.prototype.setRetriesLeft = function (retries) {
    var retriesElement = document.querySelector("#pinRetries");
    if (typeof retries === "undefined") {
        retriesElement.innerText = "?";
    } else {
        retriesElement.innerText = retries;
    }

    retriesElement = document.querySelector("#pinpadRetries");
    if (typeof retries === "undefined") {
        retriesElement.innerText = "?";
    } else {
        retriesElement.innerText = retries;
    }
};

UI.prototype.batteryStatus = function (level) {
    drawBattery("battery", level);
};